"""
This module collects together scripts to recolor images for applications such as histogram matching, color transfer etc. 
"""

import numpy as np

def image_stats(image):
    r""" Computes the mean and standard deivation of each channel in an RGB-like 2D image, (MxNx3)
    
    Parameters
    ----------
    image : (MxNx3) array
        input three channel RGB image

    Returns
    -------
    (lMean, lStd, aMean, aStd, bMean, bStd) : 6-tuple
        the mean and standard deviation of the 1st, 2nd and 3rd channels respectively 
    """
    import cv2
    # compute the mean and standard deviation of each channel
    (l, a, b) = cv2.split(image)
    (lMean, lStd) = (np.nanmean(l), np.nanstd(l))
    (aMean, aStd) = (np.nanmean(a), np.nanstd(a))
    (bMean, bStd) = (np.nanmean(b), np.nanstd(b))

    # return the color statistics
    return (lMean, lStd, aMean, aStd, bMean, bStd)

 
def color_transfer_stats(source, target, eps=1e-8):
    r""" Fast linear colour transfer between 2D 8-bit images based on mean and standard deviation colour statistics.
    
    Parameters
    ----------
    source : (MxNx3) array
        RGB image to resample color from 
    target : (MxNx3) array
        RGB image to recolor
    eps : scalar
        a small regularization number to ensure numerical stability 

    Returns
    -------
    transfer : (MxNx3) array
        the 2nd image recolored based on the statistics of the 1st image

    """
    # convert the images from the RGB to L*ab* color space, being
    # sure to utilizing the floating point data type (note: OpenCV
    # expects floats to be 32-bit, so use that instead of 64-bit)
    import cv2
    
    if target.max() > 1.0000000001:
        # ensure range in [0,1.], converts to float.  
        source = cv2.cvtColor(source, cv2.COLOR_RGB2LAB).astype("float32")
        target = cv2.cvtColor(target, cv2.COLOR_RGB2LAB).astype("float32")
    else:
        source = cv2.cvtColor(np.uint8(255*source), cv2.COLOR_RGB2LAB).astype("float32")
        target = cv2.cvtColor(np.uint8(255*target), cv2.COLOR_RGB2LAB).astype("float32")
    
    # compute color statistics for the source and target images
    (lMeanSrc, lStdSrc, aMeanSrc, aStdSrc, bMeanSrc, bStdSrc) = image_stats(source)
    (lMeanTar, lStdTar, aMeanTar, aStdTar, bMeanTar, bStdTar) = image_stats(target)

    # subtract the means from the target image
    (l, a, b) = cv2.split(target)
    l -= lMeanTar
    a -= aMeanTar
    b -= bMeanTar

    # scale by the standard deviations
    l = (lStdTar )/ (lStdSrc + eps) * l
    a = (aStdTar )/ (aStdSrc + eps) * a
    b = (bStdTar )/ (bStdSrc + eps) * b


    # add in the source mean
    l += lMeanSrc
    a += aMeanSrc
    b += bMeanSrc
    
    # clip the pixel intensities to [0, 255] if they fall outside
    # this range
    l = np.clip(l, 0, 255)
    a = np.clip(a, 0, 255)
    b = np.clip(b, 0, 255)

    # merge the channels together and convert back to the RGB color
    # space, being sure to utilize the 8-bit unsigned integer data
    # type
    transfer = cv2.merge([l, a, b])
    transfer = cv2.cvtColor(transfer.astype("uint8"), cv2.COLOR_LAB2RGB)
    
    # return the color transferred image
    return transfer
    
    
def match_color(source_img, target_img, mode='pca', eps=1e-8, source_mask=None, target_mask=None, ret_matrix=False):
    r"""Matches the colour distribution of a target image to that of a source image using a linear transform based on matrix decomposition. This effectively matches the mean and convariance of the distributions. 
    
    Images are expected to be of form (w,h,c) and can be either float in [0,1] or uint8 [0,255]. Decomposition modes are chol (Cholesky), pca (principal components) or sym for different choices of basis. The effect is slightly different in each case.
    
    Optionally image masks can be used to selectively bias the color transformation. 

    This method was popularised by the Neural style transfer paper by Gatsys et al. [1]_
    
    Parameters
    ----------
    source_img : (MxNx3) array
        RGB image to resample color from 
    target_img : (MxNx3) array
        RGB image to recolor
    mode : str
        selection of matrix decomposition 

        'pca' : str
            this is principal components or Single value decomposition of the matrix 
        'chol' : str
            this is a Cholesky matrix decomposition
        'sym' : str
            this is a symmetrizable decomposition
    eps : scalar
        a small regularization number to ensure numerical stability 
    source_mask : (MxN) array
        binary mask, specifying the select region to obtain source color statistics from 
    target_mask : (MxN) array
        binary mask, specifying the select region to recolor with color from the ``source_img`` 
    ret_matrix : bool 
        if True, returns the 3x3 matrix capturing the correlation between the colorspace of ``source_img`` and ``target_img``
    
    Returns
    -------
    matched_img : (MxNx3) array
        the 2nd image recolored based on the statistics of the 1st image

    mix_matrix : (3x3) array
        if ret_matrix==True, the 3x3 matrix capturing the correlation between the colorspace of ``source_img`` and ``target_img`` is returned as a second output

    See Also
    --------
    unwrap3D.Image_Functions.recolor.recolor_w_matrix : 
        A function to reuse the learnt mixing matrix to recolor further images without relearning. 

    References
    ----------

    .. [1] Gatys, Leon A., et al. "Controlling perceptual factors in neural style transfer." Proceedings of the IEEE conference on computer vision and pattern recognition. 2017.
    
    """
    
    if target_img.max() > 1.0000000001:
        # ensure range in [0,1.], converts to float.  
        source_img = source_img/255.
        target_img = target_img/255
    else:
        # ensure range in [0,255.]
        source_img = source_img.astype(np.float32); 
        target_img = target_img.astype(np.float32); 
    
  
    # 1. Compute the eigenvectors of the source color distribution (possibly masked)
    if source_mask is not None:
        mu_s = np.hstack([np.mean(source_img[:,:,0][source_mask==1]), np.mean(source_img[:,:,1][source_mask==1]), np.mean(source_img[:,:,2][source_mask==1])])
    else:   
        mu_s = np.hstack([np.mean(source_img[:,:,0]), np.mean(source_img[:,:,1]), np.mean(source_img[:,:,2])])
    s = source_img - mu_s # demean
    s = s.transpose(2,0,1).reshape(3,-1) # convert to (r,g,b), 3 x n_pixels
    
    if source_mask is not None:
        s = s[:, source_mask.ravel()==1] 
    Cs = s.dot(s.T) / s.shape[1] + eps * np.eye(s.shape[0]) # 3x3 covariance matrix. 
    
    # 2. Computes the eigenvectors of the target color distribution (possibly masked)    
    if target_mask is not None:
        mu_t = np.hstack([np.mean(target_img[:,:,0][target_mask==1]), np.mean(target_img[:,:,1][target_mask==1]), np.mean(target_img[:,:,2][target_mask==1])])
    else:   
        mu_t = np.hstack([np.mean(target_img[:,:,0]), np.mean(target_img[:,:,1]), np.mean(target_img[:,:,2])])
    t = target_img - mu_t
    t = t.transpose(2,0,1).reshape(3,-1)   
    
    if target_mask is not None:
        temp = t.copy()
        t = t[:, target_mask.ravel()==1] 
    Ct = t.dot(t.T) / t.shape[1] + eps * np.eye(t.shape[0]) # 3x3 covariance matrix.  

    """
    Color match the mean and covariance of the source image. 
    """    
    if mode == 'chol':
        chol_t = np.linalg.cholesky(Ct)
        chol_s = np.linalg.cholesky(Cs)
        mix_matrix = chol_s.dot(np.linalg.inv(chol_t))
        ts = mix_matrix.dot(t)
    if mode == 'pca':
        eva_t, eve_t = np.linalg.eigh(Ct)
        Qt = eve_t.dot(np.sqrt(np.diag(eva_t))).dot(eve_t.T)
        eva_s, eve_s = np.linalg.eigh(Cs)
        Qs = eve_s.dot(np.sqrt(np.diag(eva_s))).dot(eve_s.T)
        mix_matrix = Qs.dot(np.linalg.inv(Qt))
        ts = mix_matrix.dot(t)
    if mode == 'sym':
        eva_t, eve_t = np.linalg.eigh(Ct)
        Qt = eve_t.dot(np.sqrt(np.diag(eva_t))).dot(eve_t.T)
        Qt_Cs_Qt = Qt.dot(Cs).dot(Qt)
        eva_QtCsQt, eve_QtCsQt = np.linalg.eigh(Qt_Cs_Qt)
        QtCsQt = eve_QtCsQt.dot(np.sqrt(np.diag(eva_QtCsQt))).dot(eve_QtCsQt.T)
        mix_matrix = np.linalg.inv(Qt).dot(QtCsQt).dot(np.linalg.inv(Qt))
        ts = mix_matrix.dot(t)
        
    # recover the image shape. 
    if target_mask is not None:
        matched_img_flatten = np.zeros_like(temp)
        matched_img_flatten[:,target_mask.ravel()==1] = ts.copy()
    else:
        matched_img_flatten = ts.copy()
        
    matched_img = matched_img_flatten.reshape(*target_img.transpose(2,0,1).shape).transpose(1,2,0)
    matched_img += mu_s
    
    if target_mask is not None:
        rgb_mask = np.dstack([target_mask, target_mask, target_mask])
        matched_img[rgb_mask==0] = target_img[rgb_mask==0]
    
    # clip limits. 
    matched_img[matched_img>1] = 1
    matched_img[matched_img<0] = 0
    
    if ret_matrix:
        return matched_img, mix_matrix    
    else:
        return matched_img
    
    
def recolor_w_matrix(source_img, target_img, mix_matrix, source_mask=None, target_mask=None):
    r""" Matches the colour distribution of a target image to that of a source image using a linear transform based on matrix decomposition. This effectively matches the mean and convariance of the distributions and can be used instead of histogram normalization
    
    Images are expected to be of form (w,h,c) and can be either float in [0,1] or uint8 [0,255]
    Modes are chol, pca or sym for different choices of basis. The effect is slightly different in each case.
    
    Optionally image masks can be used to selectively bias the color transformation. 
    
    This method was popularised by the Neural style transfer paper by Gatsys et al. [1]_
    
    Parameters
    ----------
    source_img : (MxNx3) array
        RGB image to resample color from 
    target_img : (MxNx3) array
        RGB image to recolor
    mix_matrix : (3x3) array
        the mixing matrix capturing the covariance between the source and target img colorspaces
    source_mask : (MxN) array
        binary mask, specifying the select region to obtain source color statistics from 
    target_mask : (MxN) array
        binary mask, specifying the select region to recolor with color from the ``source_img`` 
   
    Returns
    -------
    matched_img : (MxNx3) array
        the 2nd image recolored based on the statistics of the 1st image

    See Also
    --------
    unwrap3D.Image_Functions.recolor.match_color : 
        A function to learn the mixing matrix given the source and target images  

    References
    ----------

    .. [1] Gatys, Leon A., et al. "Controlling perceptual factors in neural style transfer." Proceedings of the IEEE conference on computer vision and pattern recognition. 2017.
    
    """
    
    if target_img.max() > 1.0000000001:
        # ensure range in [0,1.], converts to float.  
        source_img = source_img/255.
        target_img = target_img/255
    else:
        # ensure range in [0,255.]
        source_img = source_img.astype(np.float32); 
        target_img = target_img.astype(np.float32); 
    
  
    # 1. Compute the eigenvectors of the source color distribution (possibly masked)
    if source_mask is not None:
        mu_s = np.hstack([np.mean(source_img[:,:,0][source_mask==1]), np.mean(source_img[:,:,1][source_mask==1]), np.mean(source_img[:,:,2][source_mask==1])])
    else:   
        mu_s = np.hstack([np.mean(source_img[:,:,0]), np.mean(source_img[:,:,1]), np.mean(source_img[:,:,2])])
    s = source_img - mu_s # demean
    s = s.transpose(2,0,1).reshape(3,-1) # convert to (r,g,b), 3 x n_pixels
    
#    if source_mask is not None:
#        s = s[:, source_mask.ravel()==1] 
#    Cs = s.dot(s.T) / s.shape[1] + eps * np.eye(s.shape[0]) # 3x3 covariance matrix. 
    
    # 2. Computes the eigenvectors of the target color distribution (possibly masked)    
    if target_mask is not None:
        mu_t = np.hstack([np.mean(target_img[:,:,0][target_mask==1]), np.mean(target_img[:,:,1][target_mask==1]), np.mean(target_img[:,:,2][target_mask==1])])
    else:   
        mu_t = np.hstack([np.mean(target_img[:,:,0]), np.mean(target_img[:,:,1]), np.mean(target_img[:,:,2])])
    t = target_img - mu_t
    t = t.transpose(2,0,1).reshape(3,-1)   
    
    temp = t.copy()
    
    # recolor the zero-meaned vector. 
    ts = mix_matrix.dot(t)
        
    
    # handles masking operations
    # recover the image shape. 
    if target_mask is not None:
        matched_img_flatten = np.zeros_like(temp)
        matched_img_flatten[:,target_mask.ravel()==1] = ts.copy()
    else:
        matched_img_flatten = ts.copy()
        
    matched_img = matched_img_flatten.reshape(*target_img.transpose(2,0,1).shape).transpose(1,2,0)
    matched_img += mu_s
    
    if target_mask is not None:
        rgb_mask = np.dstack([target_mask, target_mask, target_mask])
        matched_img[rgb_mask==0] = target_img[rgb_mask==0]
    
    # clip limits. 
    matched_img[matched_img>1] = 1
    matched_img[matched_img<0] = 0
    
    return matched_img



def calc_rgb_histogram(rgb, bins=256, I_range=(0,255)):
    r""" compute the RGB histogram of a 2D image, returning the bins and individual R, G, B histograms as a 3-tuple. 
    
    Parameters
    ----------
    rgb : (MxNx3) array
        2D RGB image 
    bins : int
        the number of intensity bins
    I_range : 2-tuple
        the (min, max) intensity range, :math:`\min\le I \le\max` to compute histograms. Intensities outside this range are not counted 
    
    Returns
    -------
    bins_ : (MxNx3) array
        the edge intensities of all bins of length ``bins`` + 1

    (r_hist,g_hist, b_hist) : 3-tuple
        individual 1-D histograms of length ``bins`` 

    """
    r_hist, bins_ = np.histogram(rgb[:,:,0].ravel(), range=I_range, bins=bins)
    g_hist, bins_ = np.histogram(rgb[:,:,1].ravel(), range=I_range, bins=bins)
    b_hist, bins_ = np.histogram(rgb[:,:,2].ravel(), range=I_range, bins=bins)
     
    return bins_, (r_hist,g_hist, b_hist)
    
    

    
    
    
    

 




def read_czifile(czi_file, squeeze=True):
    
    r""" Reads the data of a simple .czi microscope stack into a numpy array using the lightweight czifile library 

    Parameters
    ----------
    czifile : filepath
        path of the .czi file to read
    squeeze : bool
        specify whether singleton dimensions should be collapsed out

    Returns
    -------
    image_arrays : numpy array
        image stored in the .czi

    """   
    import numpy as np
    from czifile import CziFile
    
    with CziFile(czi_file) as czi:
        image_arrays = czi.asarray()
    if squeeze:
        image_arrays = np.squeeze(image_arrays)
    
    return image_arrays
    
def mkdir(directory):
    
    r"""Recursively creates all directories if they do not exist to make the folder specifed by 'directory'

    Parameters
    ----------
    directory : folderpath
        path of the folder to create
   
    """   

    import os 
    
    if not os.path.exists(directory):
        os.makedirs(directory)
        
    return []

def get_basename(pathname, ext):
   
    r"""helper to return the base file name minus the extension given an absolute or relative filepath location

    e.g. This function will return from a filepath, '../long_file_location/filename.tif', the file name minus the extension, 'filename'
    
    Parameters
    ----------
    pathname : filepath
        filepath to parse the file name from
    ext : string
        extension of the file format e.g. .tif, .png, .docx
   
    Returns
    -------
    basename : string
        the name of the file minus extension and location information
    """   

    import os 

    basename = os.path.split(pathname)[-1].split(ext)[0]

    return basename
    

def read_demons_matlab_tform( tform_file, volshape, keys=['u', 'v', 'w']):
    
    r"""helper for reading Matlab generated xyz demons deformation fields accounting for the difference in array convention.
    
    Parameters
    ----------
    tform_file : .mat (Matlab v5.4 format)
        filepath to the .mat of the saved deformation field saved from Matlab 
    volshape : (M,N,L) tuple
        shape of the original volume, the deformation fields correspond to. This shape is used to rescale any downsampled deformation fields to the original size using linear interpolation. 
    keys :  list of strings
        the variable names in the saved .mat corresponding to the x-, y-, z- direction deformation within Matlab 

    Returns
    -------
    w : (M,N,L) numpy array
        the 'x' deformation in Python tiff image reading convention 
    v : (M,N,L) numpy array
        the 'y' deformation in Python tiff image reading convention 
    u : (M,N,L) numpy array
        the 'z' deformation in Python tiff image reading convention 
    """   
    import scipy.io as spio
    import skimage.transform as sktform 
    import numpy as np 
    
    tform_obj = spio.loadmat(tform_file) # this assumes matlab v5.4 format

    u = (tform_obj[keys[0]]).astype(np.float32).transpose(2,0,1)
    v = (tform_obj[keys[1]]).astype(np.float32).transpose(2,0,1)
    w = (tform_obj[keys[2]]).astype(np.float32).transpose(2,0,1)

    scaling_factor = np.hstack(volshape).astype(np.float32) / np.hstack(u.shape)
    
    # transform this remembering to cast to float32.
    u = sktform.resize((u*scaling_factor[0]).astype(np.float32), volshape, preserve_range=True).astype(np.float32)
    v = sktform.resize((v*scaling_factor[2]).astype(np.float32), volshape, preserve_range=True).astype(np.float32)
    w = sktform.resize((w*scaling_factor[1]).astype(np.float32), volshape, preserve_range=True).astype(np.float32)

    return w,v,u


    
def save_array_to_nifti(data, savefile):

    r""" Saves a given numpy array to a nifti format using the nibabel library. The main use is for exporting volumes to annotate in ITKSnap.

    Parameters
    ----------
    data : numpy array
        input volume image
    savefile :  string
        filepath to save the output to, user should include the extension in this e.g. .nii.gz for compressed nifty

    """

    import nibabel as nib 
    import numpy as np 

    img = nib.Nifti1Image(data, np.eye(4))
    nib.save(img, savefile)  

    return []


def read_pickle(filename):
    r""" read python3 pickled .pkl files

    Parameters
    ----------
    filename : filepath
        absolute path of the file to read
   
    """   
    import pickle
    
    with open(filename, 'rb') as output:
        return pickle.load(output)


def write_pickle(savepicklefile, savedict):
    r""" write python3 pickled .pkl files - use this for objects > 4GB 

    Parameters
    ----------
    savepicklefile : filepath
        absolute path of the file to write to
    savedict : dictionary
        dictionary of variables to write 
   
    """   

    import pickle

    with open(savepicklefile, 'wb') as handle:
        pickle.dump(savedict, 
                    handle)
        
    return []
    
    
    



"""
This module generates some general default parameter settings to help get started with certain functions with many parameters 
"""

def optimesh_relaxation_config():
    r""" parameters for doing CVT relaxation of meshes using the optimesh library 
    
    see https://pypi.org/project/optimesh/
    """
    params = {}
    params['relax_method'] = 'CVT (block-diagonal)'
    params['tol'] = 1.0e-5
    params['n_iters']=20

    return params 

# Farneback 2D optical flow
def farneback2D_optical_flow():
    r""" parameters for extracting optical flow using Farneback method  
    
    see :func:`unwrap3D.Tracking.tracking.Eval_dense_optic_flow`
    """
    params = {}
    params['pyr_scale'] = 0.5
    params['levels'] = 3
    params['winsize'] = 15
    params['iterations'] = 5
    params['poly_n'] = 3
    params['poly_sigma'] = 1.2
    params['flags'] = 0 

    return params


def affine_register_matlab():
    r""" specify parameters for running Matlab imregtform to do Affine registration 
    
    see :func:`unwrap3D.Registration.registration.matlab_affine_register` for registering two timepoints and 

    see :func:`unwrap3D.Registration.registration.matlab_group_register_batch` for registering a video with switching of reference images.
    
    """
    params = {}
    params['downsample'] = [16., 8., 4.], # the scales to do registration at. [16, 8, 4, 2, 1] will continue on 2x downsampled and original resolution.
    params['modality'] = 'multimodal' # str specifying we want the multimodal registration option of imregtform
    params['type'] = 'similarity' # type of transform 
    params['view'] = None
    params['return_img'] = 1

    return params

def demons_register_matlab():
    r""" specify parameters for running Matlab imregdemons to do Demon's registration 
    
    see :func:`unwrap3D.Registration.registration.nonregister_3D_demons_matlab`
    """
    params = {}
    params['alpha'] = 0.1
    params['levels'] = [8,4,2,1]
    params['warps'] = [4,2,0,0]
    
    return params

def gradient_descent_affine_reg():
    r""" parameters for affine registration using simpleITK using Gradient Descent optimizer 
    
    see :func:`unwrap3D.Registration.registration.SITK_multiscale_affine_registration`
    """
    params = {}
    params['learningRate'] = 1
    params['numberOfIterations'] = 500
    params['convergenceMinimumValue'] = 1e-6
    params['convergenceWindowSize'] = 10 

    return params 

def evolutionary_affine_reg():
    r""" parameters for affine registration using simpleITK using Evolutionary optimizer
    
    see :func:`unwrap3D.Registration.registration.SITK_multiscale_affine_registration`
    """
    params = {}
    params['numberOfIterations'] = 100
    params['epsilon'] = 1.5e-4
    params['initialRadius'] = 6.25e-3
    params['growthFactor'] = 1.01
    params['shrinkFactor'] = -1.0 

    return params 

# demons multiscale registration in simpleITK



# from numpy import *

# def _H(y,t0=0):
#     '''
#     Step fn with step at t0
#     '''
#     h = np.zeros_like(y)
#     args = tuple([slice(0,y.shape[i]) for i in y.ndim])   

from scipy.fftpack.realtransforms import dct,idct

def smoothn(y, nS0=10,
						axis=None,
						smoothOrder=2.0,
						sd=None,
						verbose=False,
						s0=None,
						z0=None,
						isrobust=False,
						W=None,
						s=None,
						MaxIter=100,
						TolZ=1e-3, weightstr='bisquare'):
		r""" This function is translated from Matlab 

		 function [z,s,exitflag,Wtot] = smoothn(varargin)

		 SMOOTHN Robust spline smoothing for 1-D to N-D data.
		 SMOOTHN provides a fast, automatized and robust discretized smoothing
		 spline for data of any dimension.

		 see [1]_ for all the details of the algorithm 

		 Z = SMOOTHN(Y) automatically smoothes the uniformly-sampled array Y. Y
		 can be any N-D noisy array (time series, images, 3D data,...). Non
		 finite data (NaN or Inf) are treated as missing values.

		 Z = SMOOTHN(Y,S) smoothes the array Y using the smoothing parameter S.
		 S must be a real positive scalar. The larger S is, the smoother the
		 output will be. If the smoothing parameter S is omitted (see previous
		 option) or empty (i.e. S = []), it is automatically determined using
		 the generalized cross-validation (GCV) method.

		 Z = SMOOTHN(Y,W) or Z = SMOOTHN(Y,W,S) specifies a weighting array W of
		 real positive values, that must have the same size as Y. Note that a
		 nil weight corresponds to a missing value.

		 **Robust Smoothing**

			 Z = SMOOTHN(...,'robust') carries out a robust smoothing that minimizes
			 the influence of outlying data.

			 [Z,S] = SMOOTHN(...) also returns the calculated value for S so that
			 you can fine-tune the smoothing subsequently if needed.

			 An iteration process is used in the presence of weighted and/or missing
			 values. Z = SMOOTHN(...,OPTION_NAME,OPTION_VALUE) smoothes with the
			 termination parameters specified by OPTION_NAME and OPTION_VALUE. They
			 can contain the following criteria

							 *TolZ*\:       Termination tolerance on Z (default = 1e-3) TolZ must be in [0,1]

							 *MaxIter*\:    Maximum number of iterations allowed (default = 100)
							 
							 *Initial*\:    Initial value for the iterative process (default = original data)
							 
			 Syntax: [Z,...] = SMOOTHN(...,'MaxIter',500,'TolZ',1e-4,'Initial',Z0);

			 [Z,S,EXITFLAG] = SMOOTHN(...) returns a boolean value EXITFLAG that
			 describes the exit condition of SMOOTHN

							 1       SMOOTHN converged.

							 0       Maximum number of iterations was reached.

		 Parameters
		 ----------
		 y : array
				uniformly-sampled array to smooth and fill, np.ma.masked_array or np.ndarray
		 nS0 : int
				the number of samples used for cross-validation to determine optimal smoothing factor
		 axis : array
				specifies the order of the input image axis. By default assumes it is the order of the dimensions the image is passed. 
		 smoothOrder : scalar
				used in the Lambda weighting  controlling smoothness regularization in the penalized least squares process
		 sd : scalar
				used to set the weights of missing pixels if running in infilling mode. 
		 verbose : bool
				if True, print out the steps		 
		 s0 : scalar
				initial guess of s, the smoothing parameter in cross-validation 
		 z0 : array
				initial guess for z, the output, calculated with nearest neighbor if not provided
		 isrobust : bool
				if True, perform iterative reweighting for outliers
		 W : 0-1
				data weights
		 s : scalar
				smoothing parameter if not specified it is determined with GCVS (cross validation)
		 MaxIter : int
				maximum number of iterations for convergence
		 TolZ : scalar
				relative tolerance for change between iterations
		 weightstr : str
				specifies the reweighting loss for outliers if isrobust=True. One of 'cauchy', 'talworth' and bisquare'

		 Returns
		 -------
		 z : array
				the smoothed input array
		 S : scalar
				returns the calculated value for smoothing, S to allow for fine-tuning if required. 
		 exitflag : 0 or 1
				if 1, smoothn converged. if 0, the maximum number of iterations was used
		 Wtot : scalar
				return the weightings computed for the regression

		 References
		 ---------- 
		 .. [1] Garcia D, Robust smoothing of gridded data in one and higher dimensions with missing values. Computational Statistics & Data Analysis, 2010. 
		 
		 """
		import scipy.optimize.lbfgsb as lbfgsb
		import numpy.linalg
		from numpy.linalg import norm
		from scipy.fftpack.realtransforms import dct,idct
		import numpy as np
		import numpy.ma as ma

		if type(y) == ma.core.MaskedArray:  # masked array
				is_masked = True
				mask = y.mask
				y = np.array(y)
				y[mask] = 0.
				if np.any(W != None):
						W  = np.array(W)
						W[mask] = 0.
				if np.any(sd != None):
						W = np.array(1./sd**2)
						W[mask] = 0.
						sd = None
				y[mask] = np.nan
				
		if np.any(sd != None):
				sd_ = np.array(sd)
				mask = (sd > 0.)
				W = np.zeros_like(sd_)
				W[mask] = 1./sd_[mask]**2
				sd = None

		if np.any(W != None):
				W = W/W.max()

		sizy = y.shape;

		# sort axis
		if axis == None:
				axis = tuple(np.arange(y.ndim))

		noe = y.size # number of elements
		if noe<2:
				z = y
				exitflag = 0;Wtot=0
				return z,s,exitflag,Wtot
		#---
		# Smoothness parameter and weights
		#if s != None:
		#  s = []
		if np.all(W == None):
				W = np.ones(sizy);

		#if z0 == None:
		#  z0 = y.copy()

		#---
		# "Weighting function" criterion
		weightstr = weightstr.lower()
		#---
		# Weights. Zero weights are assigned to not finite values (Inf or NaN),
		# (Inf/NaN values = missing data).
		IsFinite = np.array(np.isfinite(y)).astype(bool);
		nof = IsFinite.sum() # number of finite elements
		W = W*IsFinite;
		if np.any(W<0):
				raise RuntimeError('smoothn:NegativeWeights',\
								'Weights must all be >=0')
		else:
						#W = W/np.max(W)
						pass
		#---
		# Weighted or missing data?
		isweighted = np.any(W != 1);
		#---
		# Robust smoothing?
		#isrobust
		#---
		# Automatic smoothing?
		isauto = not s;
		#---
		# DCTN and IDCTN are required
		try:
				from scipy.fftpack.realtransforms import dct,idct
		except:
				z = y
				exitflag = -1;Wtot=0
				return z,s,exitflag,Wtot

		## Creation of the Lambda tensor
		#---
		# Lambda contains the eingenvalues of the difference matrix used in this
		# penalized least squares process.
		axis = tuple(np.array(axis).flatten())
		d =  y.ndim;
		Lambda = np.zeros(sizy);
		for i in axis:
				# create a 1 x d array (so e.g. [1,1] for a 2D case
				siz0 = np.ones((1,y.ndim))[0].astype(int);
				siz0[i] = sizy[i];
				# cos(pi*(reshape(1:sizy(i),siz0)-1)/sizy(i)))
				# (arange(1,sizy[i]+1).reshape(siz0) - 1.)/sizy[i]
				Lambda = Lambda + (np.cos(np.pi*(np.arange(1,sizy[i]+1) - 1.)/sizy[i]).reshape(siz0))
				#else:
				#  Lambda = Lambda + siz0
		Lambda = -2.*(len(axis)-Lambda);
		if not isauto:
				Gamma = 1./(1+(s*abs(Lambda))**smoothOrder);

		## Upper and lower bound for the smoothness parameter
		# The average leverage (h) is by definition in [0 1]. Weak smoothing occurs
		# if h is close to 1, while over-smoothing appears when h is near 0. Upper
		# and lower bounds for h are given to avoid under- or over-smoothing. See
		# equation relating h to the smoothness parameter (Equation #12 in the
		# referenced CSDA paper).
		N = np.sum(np.array(sizy) != 1); # tensor rank of the y-array
		hMin = 1e-6; hMax = 0.99;
		# (h/n)**2 = (1 + a)/( 2 a)
		# a = 1/(2 (h/n)**2 -1) 
		# where a = sqrt(1 + 16 s)
		# (a**2 -1)/16
		try:
				sMinBnd = np.sqrt((((1+np.sqrt(1+8*hMax**(2./N)))/4./hMax**(2./N))**2-1)/16.);
				sMaxBnd = np.sqrt((((1+np.sqrt(1+8*hMin**(2./N)))/4./hMin**(2./N))**2-1)/16.);
		except:
				sMinBnd = None
				sMaxBnd = None
		## Initialize before iterating
		#---
		Wtot = W;
		#--- Initial conditions for z
		if isweighted:
				#--- With weighted/missing data
				# An initial guess is provided to ensure faster convergence. For that
				# purpose, a nearest neighbor interpolation followed by a coarse
				# smoothing are performed.
				#---
				if z0 != None: # an initial guess (z0) has been provided
								z = z0;
				else:
								z = y #InitialGuess(y,IsFinite);
								z[~IsFinite] = 0.
		else:
				z = np.zeros(sizy);
		#---
		z0 = z;
		y[~IsFinite] = 0; # arbitrary values for missing y-data
		#---
		tol = 1.;
		RobustIterativeProcess = True;
		RobustStep = 1;
		nit = 0;
		#--- Error on p. Smoothness parameter s = 10^p
		errp = 0.1;
		#opt = optimset('TolX',errp);
		#--- Relaxation factor RF: to speedup convergence
		RF = 1 + 0.75*isweighted;
		# ??
		## Main iterative process
		#---
		if isauto:
				try:
						xpost = np.array([(0.9*np.log10(sMinBnd) + np.log10(sMaxBnd)*0.1)])
				except:
						np.array([100.])
		else:
				xpost = np.array([np.log10(s)])
		while RobustIterativeProcess:
				#--- "amount" of weights (see the function GCVscore)
				aow = np.sum(Wtot)/noe; # 0 < aow <= 1
				print (aow) 
				#---
				while tol>TolZ and nit<MaxIter:
								if verbose:
										print('tol',tol,'nit',nit)
								nit = nit+1;
								DCTy = _dctND(Wtot*(y-z)+z,f=dct);
								if isauto and not np.remainder(np.log2(nit),1):
												#---
												# The generalized cross-validation (GCV) method is used.
												# We seek the smoothing parameter s that minimizes the GCV
												# score i.e. s = Argmin(GCVscore).
												# Because this process is time-consuming, it is performed from
												# time to time (when nit is a power of 2)
												#---
												# errp in here somewhere
												
												#xpost,f,d = lbfgsb.fmin_l_bfgs_b(gcv,xpost,fprime=None,factr=10.,\
												#   approx_grad=True,bounds=[(log10(sMinBnd),log10(sMaxBnd))],\
												#   args=(Lambda,aow,DCTy,IsFinite,Wtot,y,nof,noe))

												# if we have no clue what value of s to use, better span the
												# possible range to get a reasonable starting point ...
												# only need to do it once though. nS0 is teh number of samples used
												if not s0:
														ss = np.arange(nS0)*(1./(nS0-1.))*(np.log10(sMaxBnd)-np.log10(sMinBnd))+ np.log10(sMinBnd)
														g = np.zeros_like(ss)
														for i,p in enumerate(ss):
																g[i] = _gcv(p,Lambda,aow,DCTy,IsFinite,Wtot,y,nof,noe,smoothOrder)
																#print 10**p,g[i]
														xpost = [ss[g==g.min()]]
														#print '==============='
														#print nit,tol,g.min(),xpost[0],s
														#print '==============='
												else:
														xpost = [s0]
												xpost,f,d = lbfgsb.fmin_l_bfgs_b(_gcv,xpost,fprime=None,factr=1e7,\
														 approx_grad=True,bounds=[(np.log10(sMinBnd),np.log10(sMaxBnd))],\
														 args=(Lambda,aow,DCTy,IsFinite,Wtot,y,nof,noe,smoothOrder))
								s = 10**xpost[0];
								# update the value we use for the initial s estimate
								s0 = xpost[0]

								Gamma = 1./(1+(s*np.abs(Lambda))**smoothOrder);

								z = RF*_dctND(Gamma*DCTy,f=idct) + (1-RF)*z;
								# if no weighted/missing data => tol=0 (no iteration)
								tol = isweighted*norm(z0-z)/norm(z);
						 
								z0 = z; # re-initialization
				exitflag = nit<MaxIter;

				if isrobust: #-- Robust Smoothing: iteratively re-weighted process
								#--- average leverage
								h = np.sqrt(1+16.*s); 
								h = np.sqrt(1+h)/np.sqrt(2)/h; 
								h = h**N;
								#--- take robust weights into account
								Wtot = W*_RobustWeights(y-z,IsFinite,h,weightstr);
								#--- re-initialize for another iterative weighted process
								isweighted = True; tol = 1; nit = 0; 
								#---
								RobustStep = RobustStep+1;
								RobustIterativeProcess = RobustStep<3; # 3 robust steps are enough.
				else:
								RobustIterativeProcess = False; # stop the whole process

		## Warning messages
		#---
		if isauto:
				if np.abs(np.log10(s)-np.log10(sMinBnd))<errp:
								_warning('MATLAB:smoothn:SLowerBound',\
												['s = %.3f '%(s) + ': the lower bound for s '\
												+ 'has been reached. Put s as an input variable if required.'])
				elif np.abs(np.log10(s)-np.log10(sMaxBnd))<errp:
								_warning('MATLAB:smoothn:SUpperBound',\
												['s = %.3f '%(s) + ': the upper bound for s '\
												+ 'has been reached. Put s as an input variable if required.'])
				#warning('MATLAB:smoothn:MaxIter',\
				#    ['Maximum number of iterations (%d'%(MaxIter) + ') has '\
				#    + 'been exceeded. Increase MaxIter option or decrease TolZ value.'])
		return z,s,exitflag,Wtot

def _warning(s1,s2):
		print(s1)
		print(s2[0])

## GCV score
#---
#function GCVscore = gcv(p)
def _gcv(p,Lambda,aow,DCTy,IsFinite,Wtot,y,nof,noe,smoothOrder):
				import numpy as np 
				# Search the smoothing parameter s that minimizes the GCV score
				#---
				s = 10**p;
				Gamma = 1./(1+(s*np.abs(Lambda))**smoothOrder);
				#--- RSS = Residual sum-of-squares
				if aow>0.9: # aow = 1 means that all of the data are equally weighted
								# very much faster: does not require any inverse DCT
								RSS = np.linalg.norm(DCTy*(Gamma-1.))**2;
				else:
								# take account of the weights to calculate RSS:
								yhat = _dctND(Gamma*DCTy,f=idct);
								RSS = np.linalg.norm(np.sqrt(Wtot[IsFinite])*(y[IsFinite]-yhat[IsFinite]))**2;
				#---
				TrH = np.sum(Gamma);
				GCVscore = RSS/float(nof)/(1.-TrH/float(noe))**2;
				return GCVscore

## Robust weights
#function W = RobustWeights(r,I,h,wstr)
def _RobustWeights(r,I,h,wstr):
				# weights for robust smoothing.
				MAD = np.median(np.abs(r[I]-np.median(r[I]))); # median absolute deviation
				u = abs(r/(1.4826*MAD)/np.sqrt(1-h)); # studentized residuals
				if wstr == 'cauchy':
								c = 2.385; W = 1./(1+(u/c)**2); # Cauchy weights
				elif wstr == 'talworth':
								c = 2.795; W = u<c; # Talworth weights
				else:
								c = 4.685; W = (1-(u/c)**2)**2.*((u/c)<1); # bisquare weights

				W[np.isnan(W)] = 0;
				return W

## Initial Guess with weighted/missing data
#function z = InitialGuess(y,I)
def _InitialGuess(y,I):
				#-- nearest neighbor interpolation (in case of missing values)
				if any(~I):
								try:
										from scipy.ndimage.morphology import distance_transform_edt
										#if license('test','image_toolbox')
										#[z,L] = bwdist(I);
										L = distance_transform_edt(1-I)
										z = y;
										z[~I] = y[L[~I]];
								except:
										# If BWDIST does not exist, NaN values are all replaced with the
										# same scalar. The initial guess is not optimal and a warning
										# message thus appears.
										z = y;
										z[~I] = np.nanmean(y[I]);
				else:
								z = y;
				# coarse fast smoothing
				z = _dctND(z,f=dct)
				k = np.array(z.shape)
				m = np.ceil(k/10)+1
				d = []
				for i in range(len(k)):
						d.append(np.arange(m[i],k[i]))
				d = np.array(d).astype(int)
				z[d] = 0.
				z = _dctND(z,f=idct)
				return z
				#-- coarse fast smoothing using one-tenth of the DCT coefficients
				#siz = z.shape;
				#z = dct(z,norm='ortho',type=2);
				#for k in np.arange(len(z.shape)):
				#    z[ceil(siz[k]/10)+1:-1] = 0;
				#    ss = tuple(roll(array(siz),1-k))
				#    z = z.reshape(ss)
				#    z = np.roll(z.T,1)
				#z = idct(z,norm='ortho',type=2);


# NB: filter is 2*I - (np.roll(I,-1) + np.roll(I,1))
def _dctND(data,f=dct):
		nd = len(data.shape)
		if nd == 1:
				return f(data,norm='ortho',type=2)
		elif nd == 2:
				return f(f(data,norm='ortho',type=2).T,norm='ortho',type=2).T
		elif nd ==3:
				return f(f(f(data,norm='ortho',type=2,axis=0)\
																				 ,norm='ortho',type=2,axis=1)\
																				 ,norm='ortho',type=2,axis=2)
		elif nd ==4:
				return f(f(f(f(data,norm='ortho',type=2,axis=0)\
																						 ,norm='ortho',type=2,axis=1)\
																						 ,norm='ortho',type=2,axis=2)\
																						 ,norm='ortho',type=2,axis=3) 
def _peaks(n):   
		'''
		Mimic basic of matlab peaks fn
		'''
		import numpy as np 
		
		xp = np.arange(n)
		[x,y] = np.meshgrid(xp,xp)
		z = np.zeros_like(x).astype(float)
		for i in np.arange(n/5):
				x0 = np.random.random()*n
				y0 = np.random.random()*n
				sdx = np.random.random()*n/4.
				sdy = sdx
				c = np.random.random()*2 - 1.
				f = np.exp(-((x-x0)/sdx)**2-((y-y0)/sdy)**2 - (((x-x0)/sdx))*((y-y0)/sdy)*c)
				#f /= f.sum()
				f *= np.random.random()
				z += f
		return z 

				# ''
def _test1():
		 plt.figure(1)
		 plt.clf()
		 # 1-D example
		 x = np.linspace(0,100,2**8);
		 y = np.cos(x/10)+(x/50)**2 + np.random.randn(np.size(x))/10;
		 y[[70, 75, 80]] = [5.5, 5, 6];
		 z = smoothn(y)[0]; # Regular smoothing
		 zr = smoothn(y,isrobust=True)[0]; # Robust smoothing
		 plt.subplot(121)
		 plt.plot(x,y,'r.')
		 plt.plot(x,z,'k')
		 plt.title('Regular smoothing')
		 plt.subplot(122)
		 plt.plot(x,y,'r.')
		 plt.plot(x,zr,'k')
		 plt.title('Robust smoothing')
		 plt.show()

def _test2(axis=None):
		 # 2-D example
		 plt.figure(2)
		 plt.clf()
		 xp = np.arange(0,1,.02)
		 [x,y] = np.meshgrid(xp,xp);
		 f = np.exp(x+y) + np.sin((x-2*y)*3);
		 fn = f + (np.random.randn(f.size)*0.5).reshape(f.shape);
		 fs = smoothn(fn,axis=axis)[0];
		 plt.subplot(121); plt.imshow(fn,interpolation='Nearest');# axis square
		 plt.subplot(122); plt.imshow(fs,interpolation='Nearest'); # axis square

def _test3(axis=None):
		 # 2-D example with missing data
		 plt.figure(3)
		 plt.clf()
		 n = 256;
		 y0 = _peaks(n);
		 y = (y0 + np.random.random(np.shape(y0))*2 - 1.0).flatten();
		 I = np.random.permutation(range(n**2));
		 y[I[1:int(n**2*0.5)]] = np.nan; # lose 50% of data
		 y = y.reshape(y0.shape)
		 y[40:90,140:190] = np.nan; # create a hole
		 yData = y.copy()
		 
		 plt.figure()
		 plt.imshow(yData)
		 plt.show()
		 
		 z0,s,exitflag,Wtot = smoothn(yData,axis=axis); # smooth data
		 yData = y.copy()
		 z,s,exitflag,Wtot = smoothn(yData,isrobust=True,axis=axis); # smooth data
		 y = yData
		 vmin = np.min([np.min(z),np.min(z0),np.min(y),np.min(y0)])
		 vmax = np.max([np.max(z),np.max(z0),np.max(y),np.max(y0)])
		 plt.subplot(221); plt.imshow(y,interpolation='Nearest',vmin=vmin,vmax=vmax);
		 plt.title('Noisy corrupt data')
		 plt.subplot(222); plt.imshow(z0,interpolation='Nearest',vmin=vmin,vmax=vmax);
		 plt.title('Recovered data #1')
		 plt.subplot(223); plt.imshow(z,interpolation='Nearest',vmin=vmin,vmax=vmax);
		 plt.title('Recovered data #2')
		 plt.subplot(224); plt.imshow(y0,interpolation='Nearest',vmin=vmin,vmax=vmax);
		 plt.title('... compared with original data')

def _test4(i=10,step=0.2,axis=None):
		[x,y,z] = np.mgrid[-2:2:step,-2:2:step,-2:2:step]
		x = np.array(x);y=np.array(y);z=np.array(z)
		xslice = [-0.8,1]; yslice = 2; zslice = [-2,0];
		v0 = x*np.exp(-x**2-y**2-z**2)
		vn = v0 + np.random.randn(x.size).reshape(x.shape)*0.06
		v = smoothn(vn)[0];
		plt.figure(4)
		plt.clf()
		vmin = np.min([np.min(v[:,:,i]),np.min(v0[:,:,i]),np.min(vn[:,:,i])])
		vmax = np.max([np.max(v[:,:,i]),np.max(v0[:,:,i]),np.max(vn[:,:,i])])
		plt.subplot(221); plt.imshow(v0[:,:,i],interpolation='Nearest',vmin=vmin,vmax=vmax);
		plt.title('clean z=%d'%i)
		plt.subplot(223); plt.imshow(vn[:,:,i],interpolation='Nearest',vmin=vmin,vmax=vmax);
		plt.title('noisy')
		plt.subplot(224); plt.imshow(v[:,:,i],interpolation='Nearest',vmin=vmin,vmax=vmax);
		plt.title('cleaned')


def _test5():
		 t = np.linspace(0,2*np.pi,1000);
		 x = 2*np.cos(t)*(1-np.cos(t)) + np.random.randn(np.size(t))*0.1;
		 y = 2*np.sin(t)*(1-np.cos(t)) + np.random.randn(np.size(t))*0.1;
		 zx = smoothn(x)[0];
		 zy = smoothn(y)[0];
		 plt.figure(5)
		 plt.clf()
		 plt.title('Cardioid')
		 plt.plot(x,y,'r.')
		 plt.plot(zx,zy,'k')

def _test6(noise=0.05,nout=30):
		plt.figure(6)
		plt.clf()
		[x,y] = np.meshgrid(np.linspace(0,1,24),np.linspace(0,1,24))
		Vx0 = np.cos(2*np.pi*x+np.pi/2)*np.cos(2*np.pi*y);
		Vy0 = np.sin(2*np.pi*x+np.pi/2)*np.sin(2*np.pi*y);
		Vx = Vx0 + noise*np.random.randn(24,24); # adding Gaussian noise
		Vy = Vy0 + noise*np.random.randn(24,24); # adding Gaussian noise
		I = np.random.permutation(range(Vx.size))
		Vx = Vx.flatten()
		Vx[I[0:nout]] = (np.random.rand(nout)-0.5)*5; # adding outliers
		Vx = Vx.reshape(Vy.shape)
		Vy = Vy.flatten()
		Vy[I[0:nout]] = (np.random.rand(nout)-0.5)*5; # adding outliers
		Vy = Vy.reshape(Vx.shape)
		Vsx = smoothn(Vx,isrobust=True)[0];
		Vsy = smoothn(Vy,isrobust=True)[0];
		plt.subplot(131);plt.quiver(x,y,Vx,Vy,2.5)
		plt.title('Noisy')
		plt.subplot(132); plt.quiver(x,y,Vsx,Vsy) 
		plt.title('Recovered')
		plt.subplot(133); plt.quiver(x,y,Vx0,Vy0)          
		plt.title('Original')


def _sparseSVD(D):
		import scipy.sparse
		try:
				import sparsesvd
		except:
				print('bummer ... better get sparsesvd')
				exit(0)
		Ds = scipy.sparse.csc_matrix(D)
		a = sparsesvd.sparsesvd(Ds,Ds.shape[0])
		return a


def _sparseTest(n=1000):
		I = np.identity(n)
 
		# define a 'traditional' D1 matrix
		# which is a right-side difference
		# and which is *not* symmetric :-( 
		D1 = np.matrix(I - np.roll(I,1))
		# so define a symemtric version
		D1a = D1.T - D1

		U, s, Vh = scipy.linalg.svd(D1a)

		# now, get eigenvectors for D1a
		Ut,eigenvalues,Vt = _sparseSVD(D1a)
		Ut = np.matrix(Ut)


		# then, an equivalent 2nd O term would be
		D2a = D1a**2

		# show we can recover D1a
		D1a_est = Ut.T * np.diag(eigenvalues) * Ut


		# Now, because D2a (& the target D1a) are symmetric:
		D1a_est = Ut.T * np.diag(eigenvalues**0.5) * Ut


		D = 2*I - (np.roll(I,-1) + np.roll(I,1))
		a = _sparseSVD(-D)
		eigenvalues = np.matrix(a[1])
		Ut = np.matrix(a[0])
		Vt = np.matrix(a[2])
		orig = (Ut.T * np.diag(np.array(eigenvalues).flatten()) * Vt)

		Feigenvalues = np.diag(np.array(np.c_[eigenvalues,0]).flatten())
		FUt = np.c_[Ut.T,np.zeros(Ut.shape[1])]
		# confirm: FUt * Feigenvalues * FUt.T ~= D


		# m is a 1st O difference matrix
		# with careful edge conditions
		# such that m.T * m = D2
		# D2 being a 2nd O difference matrix
		m = np.matrix(np.identity(100) - np.roll(np.identity(100),1))
		m[-1,-1] = 0
		m[0,0] = 1
		a = _sparseSVD(m)
		eigenvalues = np.matrix(a[1])
		Ut = np.matrix(a[0])
		Vt = np.matrix(a[2])
		orig = (Ut.T * np.diag(np.array(eigenvalues).flatten()) * Vt)
		# Vt* Vt.T = I
		# Ut.T * Ut = I
		# ((Vt.T * (np.diag(np.array(eigenvalues).flatten())**2)) * Vt)
		# we see you get the same as m.T * m by squaring the eigenvalues

# '''
if __name__=="__main__":
				
				# from numpy import *
				import numpy as np 
				import pylab as plt 
				import scipy 
				_test1()
				_test2()
				_test3()
				_test4()
				_test5()
				_test6()
				




from ..Utility_Functions import file_io as fio
import numpy as np 
from scipy import signal


class DsiftExtractor:
	r"""
	The class that does dense sift feature extractor. See https://github.com/Yangqing/dsift-python
	Sample Usage:
		extractor = DsiftExtractor(gridSpacing,patchSize,[optional params])
		feaArr,positions = extractor.process_image(Image)
	
	Reference : 
		Y. Jia and T. Darrell. "Heavy-tailed Distances for Gradient Based Image Descriptors". NIPS 2011.
		
		Lowe, David G. "Object recognition from local scale-invariant features." Proceedings of the seventh IEEE international conference on computer vision. Vol. 2. Ieee, 1999.
	"""
	def __init__(self, gridSpacing, patchSize,
				 nrml_thres = 1.0,\
				 sigma_edge = 0.8,\
				 sift_thres = 0.2,
				 Nangles = 8,
				 Nbins = 4,
				 alpha = 9.0):
		'''
		gridSpacing: 
			the spacing for sampling dense descriptors
		patchSize: int
			the size of each sift patch
		nrml_thres: scalar
			low contrast normalization threshold
		sigma_edge: scalar
			the standard deviation for the gaussian smoothing before computing the gradient
		sift_thres: scalar
			sift thresholding (0.2 works well based on Lowe's SIFT paper)
		'''
		self.Nangles = Nangles
		self.Nbins = Nbins
		self.alpha = alpha
		self.Nsamples = Nbins**2
		self.angles = np.array(range(Nangles))*2.0*np.pi/Nangles # the thresholds of the angle histogram [0,2pi]
		
		self.gS = gridSpacing
		self.pS = patchSize
		self.nrml_thres = nrml_thres
		self.sigma = sigma_edge
		self.sift_thres = sift_thres
		# compute the weight contribution map
		sample_res = self.pS / np.double(Nbins) # spatial resolution. 
		sample_p = np.array(range(self.pS))
		sample_ph, sample_pw = np.meshgrid(sample_p,sample_p) # this is 32 x 32 (image squared?)
		sample_ph.resize(sample_ph.size)
		sample_pw.resize(sample_pw.size)
		bincenter = np.array(range(1,Nbins*2,2)) / 2.0 / Nbins * self.pS - 0.5 
#        print(bincenter)
		bincenter_h, bincenter_w = np.meshgrid(bincenter,bincenter)
#        print(bincenter_h)
#        print(bincenter_w)
		bincenter_h.resize((bincenter_h.size,1))
		bincenter_w.resize((bincenter_w.size,1))
		dist_ph = abs(sample_ph - bincenter_h)
		dist_pw = abs(sample_pw - bincenter_w)
		weights_h = dist_ph / sample_res
		weights_w = dist_pw / sample_res
		weights_h = (1-weights_h) * (weights_h <= 1)
		weights_w = (1-weights_w) * (weights_w <= 1)
		# weights is the contribution of each pixel to the corresponding bin center
		self.weights = weights_h * weights_w

	def gen_dgauss(self,sigma):
		r'''
		generates a derivative of Gaussian filter with the same specified :math:`\sigma` in both the X and Y
		directions.
		'''
		fwid = int(2*np.ceil(sigma))
		G = np.array(range(-fwid,fwid+1))**2
		G = G.reshape((G.size,1)) + G
		G = np.exp(- G / 2.0 / sigma / sigma)
		G /= np.sum(G)
		GH,GW = np.gradient(G)
		GH *= 2.0/np.sum(np.abs(GH))
		GW *= 2.0/np.sum(np.abs(GW))
		return GH,GW
		
	def process_image(self, image, positionNormalize = True,\
					   verbose = True):
		'''
		processes a single image, return the locations
		and the values of detected SIFT features.
		image: a M*N image which is a numpy 2D array. If you pass a color image, it will automatically be convertedto a grayscale image.
		positionNormalize: whether to normalize the positions to [0,1]. If False, the pixel-based positions of the top-right position of the patches is returned.
		
		Return values:
		feaArr: the feature array, each row is a feature
		positions: the positions of the features
		'''

		image = image.astype(np.double)
		if image.ndim == 3:
			# we do not deal with color images.
			image = np.mean(image,axis=2)
		# compute the grids
		H,W = image.shape
		gS = self.gS
		pS = self.pS
		remH = np.mod(H-pS, gS)
		remW = np.mod(W-pS, gS)
		offsetH = remH//2
		offsetW = remW//2
		gridH,gridW = np.meshgrid(range(offsetH,H-pS+1,gS), range(offsetW,W-pS+1,gS))
		gridH = gridH.flatten()
		gridW = gridW.flatten()
		if verbose:
			print ('Image: w {}, h {}, gs {}, ps {}, nFea {}'.\
					format(W,H,gS,pS,gridH.size))
		feaArr = self.calculate_sift_grid(image,gridH,gridW) # this is the heavy lifting. 
		feaArr = self.normalize_sift(feaArr)
		if positionNormalize:
			positions = np.vstack((gridH / np.double(H), gridW / np.double(W)))
		else:
			positions = np.vstack((gridH, gridW))
		return feaArr, positions

	def calculate_sift_grid(self,image,gridH,gridW):
		'''
		This function calculates the unnormalized sift features at equidistantly spaced control points in the image as specified by the number in height (gridH) and in width (gridW)
		It is called by process_image().
		'''
		from scipy import signal

		H,W = image.shape
		Npatches = gridH.size
		feaArr = np.zeros((Npatches, self.Nsamples*self.Nangles)) # Nsamples is the number of grid positions of the image being taken. # number of angles 
		
		# calculate gradient
		GH,GW = self.gen_dgauss(self.sigma) # this is the gradient filter for the image. 
		
		IH = signal.convolve2d(image,GH,mode='same')
		IW = signal.convolve2d(image,GW,mode='same')
		Imag = np.sqrt(IH**2+IW**2)
		Itheta = np.arctan2(IH,IW)
		Iorient = np.zeros((self.Nangles,H,W))
		
		for i in range(self.Nangles):
			Iorient[i] = Imag * np.maximum(np.cos(Itheta - self.angles[i])**self.alpha,0) # either increment the count or not. 
			#pyplot.imshow(Iorient[i])
			#pyplot.show()
		for i in range(Npatches):
			currFeature = np.zeros((self.Nangles,self.Nsamples))
			for j in range(self.Nangles):
				# this is the gaussian spatial weights in each cell. 
				currFeature[j] = np.dot(self.weights,\
						Iorient[j,gridH[i]:gridH[i]+self.pS, gridW[i]:gridW[i]+self.pS].flatten())
			feaArr[i] = currFeature.flatten()
		return feaArr

	def normalize_sift(self,feaArr):
		'''
		This function does sift feature normalization
		following David Lowe's definition (normalize length ->
		thresholding at 0.2 -> renormalize length)
		'''
		siftlen = np.sqrt(np.sum(feaArr**2,axis=1)) # this is the L2
		hcontrast = (siftlen >= self.nrml_thres)
		siftlen[siftlen < self.nrml_thres] = self.nrml_thres
		# normalize with contrast thresholding
		feaArr /= siftlen.reshape((siftlen.size,1))
		# suppress large gradients
		feaArr[feaArr>self.sift_thres] = self.sift_thres
		# renormalize high-contrast ones
		feaArr[hcontrast] /= np.sqrt(np.sum(feaArr[hcontrast]**2,axis=1)).\
				reshape((feaArr[hcontrast].shape[0],1))
		return feaArr

class SingleSiftExtractor(DsiftExtractor):
	'''
	The simple wrapper class that does feature extraction, treating
	the whole image as a local image patch.
	'''
	def __init__(self, patchSize,
				 nrml_thres = 1.0,\
				 sigma_edge = 0.8,\
				 sift_thres = 0.2,
				 Nangles = 8,
				 Nbins = 4,
				 alpha = 9.0):
		# simply call the super class __init__ with a large gridSpace
		DsiftExtractor.__init__(self, patchSize, patchSize, nrml_thres, sigma_edge, sift_thres)   
	
	def process_image(self, image):
		return DsiftExtractor.process_image(self, image, False, False)[0]

	



from ..Mesh import meshtools as meshtools
from ..Segmentation import segmentation as segmentation
from ..Unzipping import unzip as uzip 
from ..Image_Functions import image as image_fn


def uv_depth_pts3D_to_xyz_pts3D( uv_pts3D, uv_depth_params):
    r""" Linear Interpolation of the corresponding (x,y,z) coordinates given query (d,u,v) topography coordinates where the injective map (d,u,v) -> (x,y,z) is given by uv_depth_params.

    Parameters
    ----------
    uv_pts3D : (Nx3) array
        the topography, (d,u,v) coordinate at N positions for which original (x,y,z) coordinates is desired 
    uv_depth_params : (D,U,V,3) array
        the lookup table mapping the uniform (D,U,V) grid to (x,y,z) in original shape space 
    
    Returns
    -------
    xyz_pts3D : (Nx3) array
        The corresponding (x,y,z) coordinates into the space indexed by uv_depth_params.
    """

    import numpy as np 

    uv_pts3D_ = uv_pts3D.copy()
    uv_pts3D_[...,0] = np.clip(uv_pts3D[:,0], 0, uv_depth_params.shape[0]-1)
    uv_pts3D_[...,1] = np.clip(uv_pts3D[:,1], 0, uv_depth_params.shape[1]-1)
    uv_pts3D_[...,2] = np.clip(uv_pts3D[:,2], 0, uv_depth_params.shape[2]-1)

    xyz_pts3D = np.array([image_fn.map_intensity_interp3(uv_pts3D_, 
                                                    grid_shape=uv_depth_params.shape[:-1], 
                                                    I_ref=uv_depth_params[...,ch]) for ch in np.arange(uv_depth_params.shape[-1])])
    xyz_pts3D = xyz_pts3D.T.copy()
    return xyz_pts3D

def estimate_base_topography_uv_img(depth_binary):
    r""" Given a binary in topography space, this function attempts to derive a 1-to-1 height map of the basal surface by projecting straight lines in d at every (u,v) pixel position to find contiguous stretches of intracellular space. The basal surface is given by the highest d in the longest stretch of each (u,v) position 
    The resulting height map is intended to be used for downstream processing.

    Parameters
    ----------
    depth_binary : (DxUxV) array
        a binary volume in topography space where 1 indicates intracellular space and 0 is background ( extracellular ) 
    
    Returns
    -------
    heigh_func : (UxV) array
        a grayscale image with intensity = to the height coordinate in d i.e. we express height as a function of (u,v) position, height = f(u,v) 
    """

    import numpy as np 
    
    m, n = depth_binary.shape[1:]
    heigh_func = np.zeros((m,n))
    YY,XX = np.indices((m,n))

    for ii in np.arange(m)[:]:
        for jj in np.arange(n)[:]:
            # iterate over this and parse the column the longest contig. 
            data = depth_binary[:,ii,jj] > 0
            valid = np.arange(len(data))[data > 0]  # convert to index. 
            
            # break into contigs
            contigs = []
            # contigs_nearest_mesh_distance = []
            contig = []
            for jjj in np.arange(len(valid)): # iterate over the sequence. 
                if jjj == 0:
                    contig.append(valid[jjj])
                else:
                    if len(contig)>0:
                        diff = valid[jjj] - contig[-1]
                        if diff == 1:
                            contig.append(valid[jjj])
                        else:
                            # we should wrap this up. in a function now. 
                            # finish a contig. 
                            # print(contig)
                            contigs.append(contig)
                            # query_contig_pt = np.hstack([contig[-1], ii, jj])
                            # print(query_contig_pt)
                            # query_contig_pt_distance = np.min(np.linalg.norm(topography_mesh_pts-query_contig_pt[None,:], axis=-1))
                            # contigs_nearest_mesh_distance.append(query_contig_pt_distance)
                            contig=[valid[jjj]] # start a new one. 
                    else:
                        contig.append(valid[jjj]) # extend cuyrrent. 
            if len(contig) > 0 :
                contigs.append(contig)
                # query_contig_pt = np.hstack([contig[-1], ii, jj])
                # query_contig_pt_distance = np.min(np.linalg.norm(topography_mesh_pts-query_contig_pt[None,:], axis=-1))
                # contigs_nearest_mesh_distance.append(query_contig_pt_distance)
                contig = []
    

            # filter by distance then take the maximum!. 
            # contigs = [contigs[kkk] for kkk in np.arange(len(contigs)) if contigs_nearest_mesh_distance[kkk]<10.] # within a minimum threshold. 
            max_contig = contigs[np.argmax([len(cc) for cc in contigs])] # we assume we take the longest contig. 
            heigh_func[ii,jj] = max_contig[-1]

    heigh_func = np.dstack([heigh_func, YY, XX])
    return heigh_func
    

def penalized_smooth_topography_uv_img(height_func, 
                                        ds=4, 
                                        padding_multiplier=4, 
                                        method='ALS', 
                                        lam=1,
                                        p=0.25, 
                                        niter=10,
                                        uv_params=None):
    r""" Applies extended 2D asymmetric least squares regression to smooth a topography surface given as a height image, that is where the surface has been parameetrized 1-to-1 with (u,v), :math:`d=f(u,v)` 

    Parameters
    ----------
    height_func : (UxV) array
        an input topography surface given as a height image such that d=f(u,v)
    ds : int
        isotropic downsampling factor of the original image, used for imposing additional smoothness + computational efficiency  
    padding_multiplier : scalar
        this specifies the padding size as a scalar multiple, 1/padding_multiplier of the downsampled image. It is used to soft enforce spherical bounds. 
    method : str
        one of 

        'ALS' : str
            Basic asymmetric least squares algorithm, see :func:`unwrap3D.Analysis_Functions.topography.baseline_als_Laplacian`   
        'airPLS' : str
            adaptive iteratively reweighted Penalized Least Squares algorithm, see :func:`unwrap3D.Analysis_Functions.topography.baseline_airPLS2D` 
    lam : scalar
        Controls the degree of smoothness in the baseline
    p : scalar
        Controls the degree of asymmetry in the weighting. p=0.5 is the same as smoothness regularized least mean squares.
    niter : int
        The number of iterations to run the algorithm. Only a few iterations is required generally. 
    uv_params : (DxUxV,3) array
        the lookup table mapping the uniform (D,U,V) grid to (x,y,z) in original shape space. If provided the smoothness regularization will take into account the metric distortion. 
    
    Returns
    -------
    out : (UxV) array
        a smoothened output topography surface given as a height image such that d=f(u,v)
    """

    import skimage.transform as sktform
    import numpy as np 
    from sklearn.feature_extraction.image import grid_to_graph
    import scipy.sparse as spsparse

    output_shape = np.hstack(height_func.shape)
    height_binary_ds = sktform.resize(height_func, 
                                      output_shape=(output_shape//ds).astype(np.int32), 
                                      preserve_range=True); 
    # this will be used to rescale. 
    height_binary_ds_max = height_binary_ds.max()
    height_binary_ds = height_binary_ds / float(height_binary_ds_max)

    if uv_params is not None:
        uv_params_ds = sktform.resize(uv_params, 
                                      output_shape=np.hstack([(output_shape//ds).astype(np.int32), uv_params.shape[-1]]), 
                                      preserve_range=True); 
    
    # use padding to help regularize. 
    padding = (height_binary_ds.shape[0]//padding_multiplier)//2 * 2+1 # this is just to make it odd!. 

    # circular padding!. 
    height_binary_ds = np.hstack([height_binary_ds[:,-padding:], 
                                  height_binary_ds,
                                  height_binary_ds[:,:padding]])
    height_binary_ds = np.vstack([np.rot90(height_binary_ds[1:padding+1],2),
                                  height_binary_ds, 
                                  np.rot90(height_binary_ds[-1-padding:-1],2)])
    
    if uv_params is not None:
        uv_params_ds = np.hstack([uv_params_ds[:,-padding:], 
                                  uv_params_ds,
                                  uv_params_ds[:,:padding]])
        uv_params_ds = np.vstack([np.rot90(uv_params_ds[1:padding+1],2),
                                      uv_params_ds, 
                                      np.rot90(uv_params_ds[-1-padding:-1],2)])
    """
    build a graph to use the laplacian - this assumes equal weights. 
    """
    if uv_params is None:
        # build the equal weight Laplacian matrix. 
        img_graph = grid_to_graph(height_binary_ds.shape[0], height_binary_ds.shape[1])
        L = spsparse.spdiags(np.squeeze(img_graph.sum(axis=1)), 0, img_graph.shape[0], img_graph.shape[1])  - img_graph # degree - adjacency matrix. 
        L = L.tocsc()
    else:
        # build the weighted Laplacian matrix using the inverse edge length weights. 
        L = meshtools.get_inverse_distance_weight_grid_laplacian(uv_params_ds, grid_pts=uv_params_ds)
        L = L.tocsc()
    
    if method == 'ALS':
        out = baseline_als_Laplacian(height_binary_ds.ravel(), L, lam=lam, p=p, niter=niter) # because there was an issue  # seems like the best thing would be to do some smooth estimated fitting through...!
        out = out.reshape(height_binary_ds.shape)
        out = out[padding:-padding,padding:-padding].copy()
        out = sktform.resize(out, output_shape=height_func.shape) * height_binary_ds_max

    if method == 'airPLS':
        out = baseline_airPLS2D(height_binary_ds.ravel(), L, lam=lam, p=p, niter=niter)
        out = out.reshape(height_binary_ds.shape)
        out = out[padding:-padding,padding:-padding].copy()
        out = sktform.resize(out, output_shape=height_func.shape) * height_binary_ds_max

    return out 


def baseline_als_Laplacian(y, D, lam, p, niter=10):
    r""" Estimates a 1D baseline signal :math:`z=g(x_1,x_2,...,x_n)` to a 1D input signal :math:`y=f(x_1,x_2,...,x_n)` parametrized by :math:`n` dimensions  using asymmetric least squares. It can also be used for generic applications where a multidimensional image requires smoothing.
    Specifically the baseline signal, :math:`z` is the solution to the following optimization problem 

    .. math::
        z = arg\,min_z \{w(y-z)^2 + \lambda\sum(\Delta z)^2\}

    where :math:`y` is the input signal, :math:`\Delta z` is the 2nd derivative or Laplacian operator, :math:`\lambda` is the smoothing regularizer and :math:`w` is an asymmetric weighting

    .. math::
        w_i = 
        \Biggl \lbrace 
        { 
        p ,\text{ if } 
          {y_i>z_i}
        \atop 
        1-p, \text{ otherwise } 
        }

    Parameters
    ----------
    signal : 1D signal 
        The 1D signal to estimate a baseline signal. 
    D :  (NxN) sparse Laplacian matrix 
        the Laplacian matrix for the signal which captures the multidimensional structure of the signal e.g. the grid graph for a 2D or 3D image or the cotangent laplacian for a mesh. 
    lam : scalar
        Controls the degree of smoothness in the baseline
    p :  scalar
        Controls the degree of asymmetry in the weighting. p=0.5 is the same as smoothness regularized least mean squares.
    niter: int
        The number of iterations to run the algorithm. Only a few iterations is required generally. 

    Returns
    -------
    z : 1D numpy array
        the estimated 1D baseline signal

    """
    from scipy import sparse
    from scipy.sparse.linalg import spsolve
    import numpy as np 
    
    L = len(y)
    # D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    w = np.ones(L)
    for i in range(niter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        z = spsolve(Z, w*y)
        w = p * (y > z) + (1-p) * (y < z)
    return z


def baseline_airPLS2D(y, D, lam=100, p=1, niter=15):
    r""" Estimates a 1D baseline signal :math:`z=g(x_1,x_2,...,x_n)` to a 1D input signal :math:`y=f(x_1,x_2,...,x_n)` parametrized by :math:`n` dimensions  using Adaptive iteratively reweighted penalized least squares for baseline fitting. It can also be used for generic applications where a multidimensional image requires smoothing.
    Specifically the baseline signal, :math:`z` is the solution to the following optimization problem 

    .. math::
        z = arg\,min_z \{w(y-z)^2 + \lambda\sum(\Delta z)^2\}

    where :math:`y` is the input signal, :math:`\Delta z` is the 2nd derivative or Laplacian operator, :math:`\lambda` is the smoothing regularizer and :math:`w` is an asymmetric weighting

    .. math::
        w_i = 
        \Biggl \lbrace 
        { 
        0 ,\text{ if } 
          {y_i\ge z_i}
        \atop 
        e^{t(y_i-z_i)/|\textbf{d}|}, \text{ otherwise } 
        }

    where the vector :math:`\textbf{d}` consists of negative elements of the subtraction, :math:`y - z` and :math:`t` is the iteration number. 
    
    Instead of constant weights in airPLS, the weight :math:`w` is adaptively weighted for faster convergence. 

    Parameters
    ----------
    signal : 1D signal 
        The 1D signal to estimate a baseline signal. 
    D :  (NxN) sparse Laplacian matrix 
        the Laplacian matrix for the signal which captures the multidimensional structure of the signal e.g. the grid graph for a 2D or 3D image or the cotangent laplacian for a mesh. 
    lam : scalar
        Controls the degree of smoothness in the baseline
    p :  scalar
        Controls the degree of asymmetry in the weighting. p=0.5 is the same as smoothness regularized least mean squares.
    niter: int
        The number of iterations to run the algorithm. Only a few iterations is required generally. 

    Returns
    -------
    z : 1D numpy array
        the estimated 1D baseline signal

    """
    import numpy as np
    import scipy.sparse as spsparse
    from scipy.sparse import csc_matrix, eye, diags
    from scipy.sparse.linalg import spsolve

    L = len(y)
    w=np.ones(L) # equal weights initially 
    for i in range(1,niter+1):
        W = spsparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.transpose())
        z = spsolve(Z, w*y)
        d=y-z # difference between original and the estimated baseline
        # w = p * (y > z) + (1-p) * (y < z) update in the original 
        dssn=np.abs(d[d<0].sum())
        if(dssn<0.001*(abs(y)).sum() or i==niter):
            if(i==niter): print('WARING max iteration reached!')
            break
        w[d>=0]=0 # d>0 means that this point is part of a peak, so its weight is set to 0 in order to ignore it
        w[d<0]=np.exp(i*np.abs(d[d<0])/dssn)
        w[0]=np.exp(i*(d[d<0]).max()/dssn) 
        w[-1]=w[0]
    return z


def segment_topography_vol_curvature_surface(vol_curvature, 
                                             vol_binary_mask,
                                             depth_ksize=1,
                                             smooth_curvature_sigma=[1],
                                             seg_method='kmeans',
                                             n_samples=10000,
                                             n_classes=3,
                                             random_state=0,
                                             scale_feats=False):
    
    r""" Segment protrusions on the unwrapped topography using multiscale mean curvature features. 

    The multiscale comes from extracting the `mean curvature <https://en.wikipedia.org/wiki/Mean_curvature>`_, :math:`H` computed as the divergence of the normalised gradient vectors of the signed distance function, :math:`\Phi`

    .. math::
        H = -\frac{1}{2}\nabla\cdot\left( \frac{\nabla \Phi}{|\nabla \Phi|}\right)
    
    and creating a multi-feature vector concatenating the smoothed :math:`H` after Gaussian smoothing with different :math:`\sigma` as specified by smooth_curvature_sigma  

    Parameters
    ----------
    vol_curvature : (MxNxL) numpy array
        the curvature computed using the definition above using the normalised gradient vector of the signed distance transform of the binary volume 
    vol_binary_mask : (MxNxL) numpy array
        the topography binary volume from which the vol_curvature was determined form 
    depth_ksize :  scalar
        the width of a ball morphological operator for extracting a binary mask of a thin shell of thickness 2 x depth_ksize capturing the topographic volume surface to be segmented. 
    smooth_curvature_sigma: list of scalars
        The list of N :math:`\sigma`'s which the vol_curvature is smoothed with to generate a N-vector :math:`[H_{\sigma_1},H_{\sigma_2}\cdots H_{\sigma_N}]` to describe the local surface topography of a voxel  
    seg_method : str
        one of two clustering methods

        kmeans : 
            `K-Means <https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html>`_ clustering algorithm
        gmm : 
            `Gaussian mixture model <https://scikit-learn.org/stable/modules/generated/sklearn.mixture.GaussianMixture.html#sklearn.mixture.GaussianMixture>`_ algorithm with full covariances 
    n_samples : int
        the number of random sampled point to fit clustering on for computational efficiency
    n_classes :int 
        the number of clusters desired
    random_state : int
        a number to fix the random seed for reproducible clustering. 
    scale_feats : bool
        if set, the features are standard scaled prior to clustering. For mean curvature feats we find setting this seemed to make the clustering worse. 

    Returns
    -------
    depth_binary_mask, H_binary_depth_clusters : (MxNxL), (MxNxL) numpy array

        depth_binary_mask : (MxNxL) numpy array
            Binary mask of a shell of the topographic surface 
        H_binary_depth_clusters : (MxNxL) numpy array
            Integer array where background is 0 and, integers refer to the different clusters and arranged such that they reflect increasing mean curvature. 

    """
    import numpy as np 
    import skimage.morphology as skmorph
    import scipy.ndimage as ndimage
    
    # curvature only makes sense relative to a binary surface. 
    depth_binary_mask = np.logical_and(skmorph.binary_dilation(vol_binary_mask>0, skmorph.ball(depth_ksize)), 
                                       np.logical_not(skmorph.binary_erosion(vol_binary_mask>0, skmorph.ball(depth_ksize))))
    if len(smooth_curvature_sigma) == 1:
        vol_curvature_smooth = depth_binary_mask*ndimage.gaussian_filter(vol_curvature, sigma=smooth_curvature_sigma[0])

        # if vol_height is not None:
        #     vol_curvature_smooth = np.concatenate([vol_curvature_smooth[...,None], 
        #                                            vol_height[...,None]], axis=-1)
        # else:
        vol_curvature_smooth = vol_curvature_smooth[...,None] # augment. 
        # if seg_method == 'kmeans':
        #     H_binary_depth_all_clusters = segmentation.multi_level_kmeans_thresh((vol_curvature_smooth[depth_binary_mask>0])[None,None,:,None],
        #                                                                        n_classes=n_classes, n_samples=n_samples, random_state=random_state, scale=scale_feats)
        # if seg_method == 'gmm':
        #     H_binary_depth_all_clusters = segmentation.multi_level_gaussian_thresh((vol_curvature_smooth[depth_binary_mask>0])[None,None,:,None],
        #                                                                        n_classes=n_classes, n_samples=n_samples, random_state=random_state, scale=scale_feats)
    else:
        vol_curvature_smooth = np.array([depth_binary_mask*ndimage.gaussian_filter(vol_curvature, sigma=sigma) for sigma in smooth_curvature_sigma])
        vol_curvature_smooth = vol_curvature_smooth.transpose(1,2,3,0) # put this in the last dimension!. 
    
        # if vol_height is not None:
        #     vol_curvature_smooth = np.concatenate([vol_curvature_smooth, vol_height[...,None]], axis=-1)

    if seg_method == 'kmeans':
        H_binary_depth_all_clusters = segmentation.multi_level_kmeans_thresh((vol_curvature_smooth[depth_binary_mask>0])[None,None,:,:],
                                                                           n_classes=n_classes, n_samples=n_samples, random_state=random_state, scale=scale_feats)
    if seg_method == 'gmm':
        H_binary_depth_all_clusters = segmentation.multi_level_gaussian_thresh((vol_curvature_smooth[depth_binary_mask>0])[None,None,:,:],
                                                                           n_classes=n_classes, n_samples=n_samples, random_state=random_state, scale=scale_feats)
    # put these back into the volume!
    H_binary_depth_all_clusters = np.squeeze(H_binary_depth_all_clusters)

    # relabel the cluster labels in increasing curvature. 
    mean_curvature_cls = np.hstack([np.nanmean((vol_curvature_smooth[depth_binary_mask>0])[H_binary_depth_all_clusters==lab,0]) for lab in np.unique(H_binary_depth_all_clusters)])
    # conduct relabelling in the order of mean_curvature_cls.
    new_cluster_labels = np.argsort(mean_curvature_cls) 

    H_binary_depth_all_clusters_ = np.zeros_like(H_binary_depth_all_clusters)
    for new_lab, old_lab in enumerate(new_cluster_labels):
        H_binary_depth_all_clusters_[H_binary_depth_all_clusters==old_lab] = new_lab
    H_binary_depth_all_clusters = H_binary_depth_all_clusters_.copy()

    H_binary_depth_clusters = np.zeros(vol_curvature.shape, dtype=np.int32) # use int!. 
    H_binary_depth_clusters[depth_binary_mask==1] = H_binary_depth_all_clusters + 1 # add 1 as these curvature clusters are not background!. 
    
    return depth_binary_mask, H_binary_depth_clusters
    
    
def remove_topography_segment_objects_binary( vol_clusters, minsize=100, uv_params_depth=None):
    r""" Removes small connected components in the binary image input either by the number of voxels or if provided the mapping to the original space, on the apparent number of voxels after geometric correction.    
    
    Parameters
    ----------
    vol_clusters : (D,U,V) array
        a binary volume
    minsize : scalar
        the minimum size, any connected components less than this is removed
    uv_params_depth :  (D,U,V,3) array
        the lookup table mapping the uniform (D,U,V) grid to (x,y,z) in original shape space, if provided the minsize is computed for (x,y,z) space not the current (d,u,v) space 
    
    Returns
    -------
    vol_clusters_new : (D,U,V) array
        a binary volume where size of connected components with > minsize removed. 
        
    """
    import numpy as np 
    import skimage.measure as skmeasure
    
    vol_clusters_new = vol_clusters.copy()
    vol_clusters_label = skmeasure.label(vol_clusters) # run connected component analysis 
    
    # measure the properties 
    props = skmeasure.regionprops(vol_clusters_label)
    
    if uv_params_depth is None:
        props_size = np.hstack([re.area for re in props]) # min_size = 100
        remove_labels = np.setdiff1d(vol_clusters_label, 0)[props_size<minsize] # should this be applied in 3D ? or just in the image? 
        
        for lab in remove_labels:
            vol_clusters_new[vol_clusters_label==lab] = 0 # set to 0 / bg
    else:
        unique_labels = np.setdiff1d(np.unique(vol_clusters_label), 0)
        if len(unique_labels) > 0: 
                                     
            # we are going to compute the actual volume using differential calculus. 
            dV_dd = np.gradient(uv_params_depth, axis=0)
            dV_dy = np.gradient(uv_params_depth, axis=1)
            dV_dx = np.gradient(uv_params_depth, axis=2)
            
            dVol = np.abs(np.nansum(np.cross(dV_dx, dV_dy, axis=-1) * dV_dd, axis=-1))
            print(dVol.shape)
            
            remove_labels = []
            for lab in unique_labels:
                vol_mask = vol_clusters_label==lab
                vol_mask_area = np.nansum(dVol[vol_mask>0])
                if vol_mask_area < minsize:
                    remove_labels.append(lab)
        
            for lab in remove_labels:
                vol_clusters_new[vol_clusters_label==lab] = 0 # set to 0 / bg
                
    return vol_clusters_new
    
    
def prop_labels_watershed_depth_slices(topo_depth_clusters, depth_binary, rev_order=True, expand_labels_2d=2):
    r""" Propagate semantic labels volumetrically from surface labels in topography space using marker-seeded watershed slice-by-slice with markers seeded from surface and from top to bottom (or bottom to top).

    This function is used to propagate surface labels into the volume so as to obtain realistic volumizations of protrusion instance segmentations when mapped back to (x.y.z) space.

    Parameters
    ----------
    topo_depth_clusters : (MxNxL) numpy array
        integer labelled volume where background voxels = 0 and unique integers > 0 represent unique connected objects.  
    depth_binary : (MxNxL) numpy array
        the topography binary volume defining the voxels that needs semantic labeling
    rev_order :  bool
        if True, reverses the scan direction to go from top to bottom instead of bottom to top (default) 
    expand_labels_2d: int
        a preprocessing expansion of input topo_depth_clusters, to better guide the in-plane watershed propagation.            
    
    Returns
    -------
    topo_depth_clusters_ : (MxNxL) numpy array
        new integer labelled volume where background voxels = 0 and unique integers > 0 represent unique connected objects. 
    """

    import numpy as np 
    import scipy.ndimage as ndimage 
    import skimage.segmentation as sksegmentation

    D = len(topo_depth_clusters)

    if rev_order == True:
        topo_depth_clusters_ = topo_depth_clusters[::-1].copy()
        depth_binary_ = depth_binary[::-1].copy() # might want to keep as 0 labels? # how to prevent flow? # we can use piecewise cuts? to approx? 
    else:
        topo_depth_clusters_ = topo_depth_clusters.copy()
        depth_binary_ = depth_binary.copy() # might want to keep as 0 labels? # how to prevent flow? # we can use piecewise cuts? to approx? 
    
    for dd in np.arange(D):
        dtform_2D = ndimage.distance_transform_edt(depth_binary_[dd]>0)
        if dd == 0: 
            label_dd = topo_depth_clusters_[dd].copy()
            label_dd_watershed = sksegmentation.watershed(-dtform_2D, 
                                                          markers=sksegmentation.expand_labels(label_dd, distance=expand_labels_2d), 
                                                          mask=depth_binary_[dd]>0)  # labels = watershed(-distance, markers, mask=image)
            topo_depth_clusters_[dd] = label_dd_watershed.copy()
        # if dd>0:
        else:
            label_dd = topo_depth_clusters_[dd].copy()
            joint_mask = np.logical_and(depth_binary_[dd]>0, depth_binary_[dd-1]>0) # get the join
            # same time there has to be a value.
            joint_mask = np.logical_and(joint_mask, topo_depth_clusters_[dd-1]>0)
            
            if np.sum(joint_mask) > 0: 
                label_dd[joint_mask] = (topo_depth_clusters_[dd-1][joint_mask]).copy() # then copy the vals. 
            
            label_dd_watershed = sksegmentation.watershed(-dtform_2D, 
                                                          markers=sksegmentation.expand_labels(label_dd, distance=expand_labels_2d), 
                                                          mask=depth_binary_[dd]>0)  # labels = watershed(-distance, markers, mask=image)
            topo_depth_clusters_[dd] = label_dd_watershed.copy()
    
    if rev_order:
        topo_depth_clusters_ = topo_depth_clusters_[::-1].copy()

    return topo_depth_clusters_
    

def inpaint_topographical_height_image(vol_labels_binary, 
                                       pre_smooth_ksize=1,
                                       post_smooth_ksize=3,
                                       background_height_thresh=None,
                                       inpaint_method='Telea',
                                       inpaint_radius=1,
                                       spherical_pad=True):
    r""" Given a topographical binary where 1 at (d,u,v) denotes the cell describe the cell surface as a height map where :math:`d=f(u,v)` and 'holes' are represented with :math:`d\le h_{thresh}`, where :math:`h_{thresh}` is a minimal height threshold, use image inpainting to 'infill' the holes to obtain a complete surface.

    Parameters
    ----------
    vol_labels_binary : (MxNxL) numpy array
        unwrapped topographic binary volume  
    pre_smooth_ksize : (MxNxL) numpy array
        gaussian :math:`\sigma` for presmoothing the height image 
    post_smooth_ksize :  scalar
        gaussian :math:`\sigma` for postsmoothing the inpainted height image 
    background_height_thresh: scalar
        all (u,v) pixels with height less than the specified threshold (:math:`d\le h_{thresh}`) is marked as 'holes' for inpainting 
    inpaint_method : str
        one of two classical inpainting methods implemented in `OpenCV <https://docs.opencv.org/3.4/df/d3d/tutorial_py_inpainting.html>`_. 

        'Telea' : 
            Uses Fast Marching method of `Telea et al. <https://docs.opencv.org/3.4/d7/d8b/group__photo__inpaint.html#gga8c5f15883bd34d2537cb56526df2b5d6a892824c38e258feb5e72f308a358d52e>`_.
        'NS' : 
            Uses Navier-Stokes method of `Bertalmio et al. <https://docs.opencv.org/3.4/d7/d8b/group__photo__inpaint.html#gga8c5f15883bd34d2537cb56526df2b5d6a892824c38e258feb5e72f308a358d52e>`_. 
    spherical_pad : int
        the number of pixels to spherically pad, to soft-mimic the closed spherical boundary conditions in original (x,y,z) space 
   
    Returns
    -------
    infill_height : (UxV) numpy array
        Inpainted height image representing the hole-completed topographic surface

    (background_height, infill_mask) : 

        background_height : (UxV) numpy array
            Height image of the input surface 
        infill_mask : (UxV) numpy array
            Binary image where 1 denotes the region to be infilled

    """

    import numpy as np 
    import cv2
    import skimage.morphology as skmorph 
    import scipy.ndimage as ndimage
    
    background_height_coords = np.argwhere(vol_labels_binary>0)
    background_height = np.zeros(vol_labels_binary.shape[1:]) # build the image. 
    background_height[background_height_coords[:,1], 
                      background_height_coords[:,2]] = background_height_coords[:,0] # assuming the 1st = height. 
    
    # mark the infill mask. 
    if pre_smooth_ksize is not None:
        if background_height_thresh is None:
            infill_mask = skmorph.binary_dilation(background_height<1, skmorph.disk(pre_smooth_ksize))
        else:
            infill_mask = skmorph.binary_dilation(background_height<background_height_thresh, skmorph.disk(pre_smooth_ksize))
    
    # for rescaling purposes. 
    background_height_max = background_height.max()
    
    if inpaint_method == 'Telea':
        infill_height = cv2.inpaint(np.uint8(255*background_height/background_height_max), np.uint8(255*infill_mask), inpaint_radius, cv2.INPAINT_TELEA)
    if inpaint_method == 'NS':
        infill_height = cv2.inpaint(np.uint8(255*background_height/background_height_max), np.uint8(255*infill_mask), inpaint_radius, cv2.INPAINT_NS)
    
    infill_height = infill_height/255. * background_height_max

    if spherical_pad:
        # we need to ensure that things connect linearly # a quick fix is to set left equal to right and the top and bottom to the mean with minimal changes.  
        infill_height_ = infill_height.copy()
        infill_height_[0,:] = np.nanmean(infill_height[0,:])
        infill_height_[-1,:] =  np.nanmean(infill_height[-1,:])
        infill_height_[:,-1] = infill_height_[:,0].copy()
        infill_height = infill_height_.copy()

    if post_smooth_ksize is not None:
        infill_height = ndimage.gaussian_filter(infill_height, sigma=post_smooth_ksize) # maybe a regulariser is a better smoother... 

    return infill_height, (background_height, infill_mask) 


def mask_volume_dense_propped_labels_with_height_image(vol_labels, height_map, 
                                                        ksize=1, 
                                                        min_size=500, 
                                                        connectivity=2,
                                                        keep_largest=False):
    r""" This function uses a hole-completed basal surface such as that from :func:`unwrap3D.Analysis_Functions.topography.inpaint_topographical_height_image` to isolate only the protrusions from a volume-dense volumization such as the output from :func:`unwrap3D.Analysis_Functions.topography.prop_labels_watershed_depth_slices`

    Parameters
    ----------
    vol_labels : (MxNxL) numpy array
        integer labelled volume where background voxels = 0 and unique integers > 0 represent unique connected objects. 
    height_map : (MxN) numpy array
        height image at every (u,v) pixel position specifying the basal surface :math:`d=f(u,v)` such that all voxels in vol_labels with coordinates :math:`d_{uv}\le d` will be set to a background label of 0
    ksize :  int
        Optional morphological dilation of the binary surface specified by height map with a ball kernel of radius ksize. Can be used to control the extent of exclusion of the basal surface. 
        Set this parameter to None to not do morphological processing.
    min_size: scalar
        The minimum size of connected components to keep after masking
    connectivity : int
        specifies the connectivity of voxel neighbors. If 1, the 6-connected neighborhood, if 2, the full 26-connected neighborhood including the diagonals 
    keep_largest : bool
        if True, for labels that end up disconnected after the masking, retain only the largest connected region. 

    Returns
    -------
    vol_labels_new : (MxNxL) numpy array
        the non excluded integer labelled volume where background voxels = 0 and unique integers > 0 represent unique connected objects. 
    infill_height_mask : (MxN) numpy array
        the exclusion volume specified by height_map after optional ksize morphological dilation.

    """

    import numpy as np 
    import skimage.morphology as skmorph 

    # use the height to mask out all invalid volumes... 
    # depth_height = depth_mesh.vertices[:,0].reshape(H_binary_.shape[1:])
    ZZ, _, _ = np.indices(vol_labels.shape)
    if ksize is not None:
        infill_height_mask = np.logical_not(skmorph.binary_dilation(np.logical_not(ZZ >= height_map), skmorph.ball(ksize)))
    else:
        infill_height_mask = ZZ >= height_map
    
    # apply mask. 
    vol_labels_new = vol_labels*infill_height_mask
    # postprocess - keep only the largest component in each cluster label!   
    keep_mask = skmorph.remove_small_objects(vol_labels_new>0, min_size=min_size, connectivity=connectivity)
    vol_labels_new = vol_labels_new*keep_mask # apply the mask. 

    if keep_largest:
        # suppressing this should be faster. 
        # for each label we should only have 1 connected component! 
        vol_labels_new = segmentation.largest_component_vol_labels(vol_labels_new, connectivity=connectivity)

    return vol_labels_new, infill_height_mask
    






import numpy as np 

def pad_vol_2_size(im, shape, fill_method='constant', fill_value=0):
	r""" Utility function to center and copy the given volumetric image into a larger volume.  
	
	Parameters
	----------
	im : (MxNxL) array
		input volume to copy
	shape : (m,n,l) array
		new volume shape
	fill_method : str
		method of specifying the missing pixels, one of 

		'constant' : str
			fill with a constant value specifed by the additional argument ``fill_value``
		'mean' : str
			fill with the mean value of ``im``
		'median' : str
			fill with the median value of ``im``
	
	fill_value : scalar
		the constant value to initialise the new array if ``fill_method=='constant'``

	Returns
	-------
	new_im : (MxNxL) array
		output volume the same shape as that specified by ``shape``
	"""
	import numpy as np 

	shift_1 = (shape[0] - im.shape[0]) // 2
	shift_2 = (shape[1] - im.shape[1]) // 2
	shift_3 = (shape[2] - im.shape[2]) // 2
	
	new_im = np.zeros(shape, dtype=np.uint8)
	new_im[shift_1:shift_1+im.shape[0],
		   shift_2:shift_2+im.shape[1],
		   shift_3:shift_3+im.shape[2]] = im.copy()

	return new_im


def mean_vol_img(vol_img_list, target_shape, fill_method='constant', fill_value=0):
	r""" Utility function to find the mean volume image given a list of volumetric image files. The volumes will automatically be padded to the specified target shape. 
	
	Parameters
	----------
	vol_img_list : array of filepaths
		list of volumetric image filepaths to compute the mean image for 
	target_shape : (m,n,l) array
		desired volume shape, must be at least the size of the largest volume shape. 
	fill_method : str
		method of specifying the missing pixels, one of 

		'constant' : str
			fill with a constant value specifed by the additional argument ``fill_value``
		'mean' : str
			fill with the mean value of ``im``
		'median' : str
			fill with the median value of ``im``
	
	fill_value : scalar
		the constant value to initialise the new array if ``fill_method=='constant'``

	Returns
	-------
	mean_vol : (MxNxL) array
		mean volume with the same datatype as that of the input volumes
	"""
	import skimage.io as skio 
	# work in float but convert to ubytes for allocation? 
	mean_vol = np.zeros(target_shape) 
	n_imgs = len(vol_img_list)

	for v in vol_img_list:
		im = skio.imread(v)
		im = pad_vol_2_size(im, target_shape, fill_method=fill_method, fill_value=fill_value) 
		mean_vol += im/float(n_imgs)
	
	mean_vol = mean_vol.astype(im.dtype)
	return mean_vol # cast to the same type


def max_vol_img(vol_img_list, target_shape, fill_method='constant', fill_value=0):
	r""" Utility function to find the maximum intensty volume image given a list of volumetric image files. The volumes will automatically be padded to the specified target shape. 
	
	Parameters
	----------
	vol_img_list : array of filepaths
		list of volumetric image filepaths to compute the mean image for 
	target_shape : (m,n,l) array
		desired volume shape, must be at least the size of the largest volume shape. 
	fill_method : str
		method of specifying the missing pixels, one of 

		'constant' : str
			fill with a constant value specifed by the additional argument ``fill_value``
		'mean' : str
			fill with the mean value of ``im``
		'median' : str
			fill with the median value of ``im``
	
	fill_value : scalar
		the constant value to initialise the new array if ``fill_method=='constant'``

	Returns
	-------
	max_vol : (MxNxL) array
		maximum intensity volume with the same datatype as that of the input volumes
	"""
	import skimage.io as skio  
	max_vol = np.zeros(target_shape) 

	for v in vol_img_list:
		im = skimage.imread(v)
		im = pad_vol_2_size(im, target_shape, fill_method=fill_method, fill_value=fill_value) 
		max_vol = np.maximum(max_vol, im)       
		
	return max_vol


def imadjust(vol, p1, p2): 
	r""" A crucial preprocessing for many applications is to contrast stretch the original microscopy volume intensities. This is done here through percentiles similar to Matlabs imadjust function.

	Parameters
	----------
	vol : nd array
		arbitrary n-dimensional image
	p1 : scalar
		lower percentile of intensity, specified as a number in the range 0-100. All intensities in a lower percentile than p1 will be mapped to 0.
	p2 : scalar
		upper percentile of intensity, specified as a number in the range 0-100. All intensities in the upper percentile than p2 will be mapped to the maximum value of the data type.
	
	Returns
	-------
	vol_rescale : nd array
		image of the same dimensions as the input with contrast enchanced

	"""
	import numpy as np 
	from skimage.exposure import rescale_intensity
	# this is based on contrast stretching and is used by many of the biological image processing algorithms.
	p1_, p2_ = np.percentile(vol, (p1,p2))
	vol_rescale = rescale_intensity(vol, in_range=(p1_,p2_))
	return vol_rescale

def normalize_mi_ma(x, mi, ma, clip=False, eps=1e-20, dtype=np.float32):
	r""" Linear image intensity rescaling based on the given minimum and maximum intensities, with optional clipping
	
	The new intensities will be 

	.. math::
		x_{new} = \frac{x-\text{mi}}{\text{ma}-\text{mi} + \text{eps}}
		
	Parameters
	----------
	x : nd array
		arbitrary n-dimensional image
	mi : scalar
		minimum intensity value 
	ma : scalar
		maximum intensity value 
	clip : bool
		whether to clip the transformed image intensities to [0,1] 
	eps : scalar
		small number for numerical stability 
	dtype : np.dtype
		if not None, casts the input x into the specified dtype 

	Returns
	-------
	x : nd array
		normalized output image of the same dimensionality as the input image

	"""
	if dtype is not None:
		x   = x.astype(dtype,copy=False)
		mi  = dtype(mi) if np.isscalar(mi) else mi.astype(dtype,copy=False)
		ma  = dtype(ma) if np.isscalar(ma) else ma.astype(dtype,copy=False)
		eps = dtype(eps)
	try:
		import numexpr
		x = numexpr.evaluate("(x - mi) / ( ma - mi + eps )")
	except ImportError:
		x =                   (x - mi) / ( ma - mi + eps )
	if clip:
		x = np.clip(x,0,1)
	return x

def normalize(x, pmin=2, pmax=99.8, axis=None, clip=False, eps=1e-20, dtype=np.float32):
	r""" Percentile-based image normalization, same as :func:`unwrap3D.Image_Functions.image.imadust` but exposes more flexibility in terms of axis selection and optional clipping. 

	Parameters
	----------
	x : nd array
		arbitrary n-dimensional image
	pmin : scalar
		lower percentile of intensity, specified as a number in the range 0-100. All intensities in a lower percentile than p1 will be mapped to 0.
	pmax : scalar
		upper percentile of intensity, specified as a number in the range 0-100. All intensities in the upper percentile than p2 will be mapped to the maximum value of the data type.
	axis : int
		if not None, compute the percentiles restricted to the specified axis, otherwise they will be based on the statistics of the whole image.
	clip : bool
		whether to clip to the maximum and minimum of [0,1] for float32 or float64 data types.    
	eps : scalar
		small number for numerical stability 
	dtype : np.dtype
		if not None, casts the input x into the specified dtype 

	Returns
	-------
	x_out : nd array
		normalized output image of the same dimensionality as the input image

	"""
	mi = np.percentile(x,pmin,axis=axis,keepdims=True)
	ma = np.percentile(x,pmax,axis=axis,keepdims=True)

	x_out = normalize_mi_ma(x, mi, ma, clip=clip, eps=eps, dtype=dtype)
	return x_out


def std_norm(im, mean=None, factor=1, cutoff=4, eps=1e-8):
	r""" Contrast normalization using the mean and standard deviation with optional clipping. 

	The intensities are transformed relative to mean :math:`\mu` and standard deviation :math:`\sigma`. The transformed pixel intensities can then be effectively gated based on properties of a Gaussian distribution. Compared to percentile normalisation this methods is easier to tune.

	.. math::
		I_{new} = \frac{I-\mu}{\text{factor}\cdot\sigma + \text{eps}}
		
	The extent of enhancement is effectively adjusted through ``factor``. If clipping is specified through ``cutoff`` the positive intensities up to ``cutoff`` is retained

	.. math::
		I_{final} &= \text{clip}(I_{new}, 0, \text{cutoff}) \\
		I_{final} &= \frac{I_{final}-\min(I_{final})}{\max(I_{final})-\min(I_{final}) + \text{eps}}


	Parameters
	----------
	im : nd array
		arbitrary n-dimensional image
	mean : scalar
		The mean intensity to subtract off. if None, it is computed as the global mean of `im`. 
	factor : scalar
		adjusts the extent of contrast enhancment. 
	cutoff : int
		if not None, clip the transformed intensities to the range [0, cutoff] and then rescale the result to [0,1]
	eps : scalar
		small number for numerical stability 
	
	Returns
	-------
	im_ : nd array
		normalized output image of the same dimensionality as the input image

	"""
	if mean is None:
		im_ = im - im.mean()
	else:
		im_ = im - mean
	im_ = im_/(factor*im.std() + eps)

	if cutoff is not None:
		im_ = np.clip(im_, 0, cutoff) # implement clipping. 
		im_ = (im_ - im_.min()) / (im_.max() - im_.min() + eps) # rescale. 
	
	return im_ 


def map_intensity_interp2(query_pts, 
						  grid_shape, 
						  I_ref, 
						  method='linear', 
						  cast_uint8=False, 
						  s=0,
						  fill_value=0):
	r""" Interpolate a 2D image at specified coordinate points. 

	Parameters
	----------
	query_pts : (n_points,2) array
		array of (y,x) image coordinates to interpolate intensites at from ``I_ref``
	grid_shape : (M,N) tuple
		shape of the image to get intensities from 
	I_ref : (M,N) array
		2D image to interpolate intensities from 
	method : str
		Interpolating algorithm. Either 'spline' to use scipy.interpolate.RectBivariateSpline or any other e.g. 'linear' from scipy.interpolate.RegularGridInterpolator
	cast_uint8 : bool
		Whether to return the result as a uint8 image. 
	s : scalar
		if method='spline', s=0 is a interpolating spline else s>0 is a smoothing spline.
	fill_value : scalar
		specifies the imputation value if any of query_pts needs to extrapolated. 
	
	Returns
	-------
	I_query : (n_points,) array
		the intensities of ``I_ref`` interpolated at the specified ``query_pts`` 

	"""
	import numpy as np 
	from scipy.interpolate import RectBivariateSpline, RegularGridInterpolator 
	
	if method == 'spline':
		spl = RectBivariateSpline(np.arange(grid_shape[0]), 
								  np.arange(grid_shape[1]), 
								  I_ref,
								  s=s)
		I_query = spl.ev(query_pts[...,0], 
						 query_pts[...,1])
	else:
		spl = RegularGridInterpolator((np.arange(grid_shape[0]), 
									   np.arange(grid_shape[1])), 
									   I_ref, method=method, bounds_error=False, 
									   fill_value=fill_value)
		I_query = spl((query_pts[...,0], 
					   query_pts[...,1]))

	if cast_uint8:
		I_query = np.uint8(I_query)
	
	return I_query

def scattered_interp2d(pts2D, pts2Dvals, gridsize, method='linear', cast_uint8=False, fill_value=None):
	r""" Given an unordered set of 2D points with associated intensities, use Delaunay triangulation to interpolate a 2D image of the given size. Grid points outside of the convex hull of the points will be set to the given ``fill_value`` 

	Parameters
	----------
	pts2D : (n_points,2)
		array of (y,x) image coordinates where intensites are known 
	pts2Dvals : (n_points,)
		array of corresponding image intensities at ``pts2D``
	gridsize : (M,N) tupe
		shape of the image to get intensities
	method : str
		one of the interpolatiion methods in ``scipy.interpolate.griddata``
	cast_uint8 : bool
		Whether to return the result as a uint8 image. 
	fill_value :
		specifies the imputation value if any of query_pts needs to extrapolated. 

	Returns
	-------
	img : (M,N)
		the interpolated dense image intensities of the given unordered 2D points
	
	"""
	import scipy.interpolate as scinterpolate
	import numpy as np 

	if fill_value is None:
		fill_value = np.nan

	YY, XX = np.indices(gridsize)

	img = scinterpolate.griddata(pts2D, 
								 pts2Dvals, 
								 (YY, XX), method=method, fill_value=fill_value)
	
	return img

def map_intensity_interp3(query_pts, 
							grid_shape, 
							I_ref, 
							method='linear', 
							cast_uint8=False,
							fill_value=0):
	r""" Interpolate a 3D volume image at specified coordinate points. 

	Parameters
	----------
	query_pts : (n_points,3) array
		array of (x,y,z) image coordinates to interpolate intensites at from ``I_ref``
	grid_shape : (M,N,L) tuple
		shape of the volume image to get intensities from 
	I_ref : (M,N,L) array
		3D image to interpolate intensities from 
	method : str
		Interpolating order. One of those e.g. 'linear' from scipy.interpolate.RegularGridInterpolator
	cast_uint8 : bool
		Whether to return the result as a uint8 image. 
	fill_value : scalar
		specifies the imputation value if any of query_pts needs to extrapolated. 
	
	Returns
	-------
	I_query : (n_points,) array
		the intensities of ``I_ref`` interpolated at the specified ``query_pts`` 

	"""
	# interpolate instead of discretising to avoid artifact.
	from scipy.interpolate import RegularGridInterpolator #, RBFInterpolator
	# from scipy.interpolate import griddata
	from scipy import ndimage
	import numpy as np 
	
	spl_3 = RegularGridInterpolator((np.arange(grid_shape[0]), 
									 np.arange(grid_shape[1]), 
									 np.arange(grid_shape[2])), 
									 I_ref, method=method, bounds_error=False, 
									 fill_value=fill_value)

	I_query = spl_3((query_pts[...,0], 
					  query_pts[...,1],
					  query_pts[...,2]))
	
	if cast_uint8:
		I_query = np.uint8(I_query)
	
	return I_query


def guidedfilter(I, G, radius, eps):
    """ n-dimensional guided filter. 
    
    see: K.He, J.Sun, and X.Tang. Guided Image Filtering. TPAMI'12.
    
    Parameters
	----------
	I : n-d array
		the input floating point image (0-1) to be filtered
	G : n-d array
		the guide floating point image (0-1) to filter I with
	radius : int
		the local radius determining the extent of influence by G.
	eps : float
		regularization parameter, the smaller the value, the more the image is filtered by G. 
	
	Returns
	-------
	q : n-d array
		the filtered image of I using the guide image G of same dimensions
    
    """
    import numpy as np 
    import scipy.ndimage as ndimage 
    
    # step 1
    meanI  = ndimage.uniform_filter(G, size=radius)
    meanp  = ndimage.uniform_filter(I, size=radius)
    corrI  = ndimage.uniform_filter(G*G, size=radius)
    corrIp = ndimage.uniform_filter(I*G, size=radius)
    
    # step 2
    varI   = corrI - meanI * meanI
    covIp  = corrIp - meanI * meanp
    
    # step 3
    a      = covIp / (varI + eps)
    b      = meanp - a * meanI
    # step 4
    meana  = ndimage.uniform_filter(a, size=eps)
    meanb  = ndimage.uniform_filter(b, size=eps)
    # step 5
    q = meana * G + meanb

    return q
